import unittest

import pulsar
from pulsar.utils.httpurl import iteritems, itervalues
from pulsar.apps.test.result import Plugin


__all__ = ['WrapTest',
           'TestOption',
           'TestPlugin']


def as_test_setting(setting):
    setting.app = 'test'
    setting.section = "Test"
    return setting


class TestOption(pulsar.Setting):
    virtual = True
    app = 'test'
    section = "Test"


class WrapTest(object):
    '''Wrap an underlying test case'''
    def __init__(self, test):
        self.test = test
        setattr(self, test._testMethodName, self._call)
        self.testMethod = getattr(test, test._testMethodName)

    def __str__(self):
        return self.test._testMethodName
    __repr__ = __str__
    
    @property
    def original_test(self):
        if isinstance(self.test, WrapTest):
            return self.test.original_test
        else:
            return self.test
        
    def set_test_attribute(self, name, value):
        setattr(self.original_test, name, value)
        
    def __getattr__(self, name):
        return getattr(self.original_test, name)
        
    def _call(self):
        # This is the actual function to implement
        return self.testMethod()
    
    
class TestPluginMeta(type):
    
    def __new__(cls, name, bases, attrs):
        settings = {}
        for base in bases:
            if isinstance(base, TestPluginMeta):
                settings.update(base.config.settings)
        for key, setting in list(iteritems(attrs)):
            if isinstance(setting, pulsar.Setting):
                attrs.pop(key)
                setting.name = setting.name or key.lower()
                settings[setting.name] = as_test_setting(setting)
        if not attrs.pop('virtual', False):
            setting_name = attrs.pop('name', name).lower()
            if setting_name:
                def_flag = '--%s' % setting_name.replace(' ','-').replace('_','-')
                action = attrs.pop('action', None)
                type = attrs.pop('type', None)
                default = attrs.pop('default', None)
                validator = attrs.pop('validator', None)
                nargs = attrs.pop('nargs', None)
                if validator is None and default is None and type is None\
                    and nargs is None:
                    if action is None or action == 'store_true':
                        action = 'store_true'
                        default = False
                        validator = pulsar.validate_bool
                    elif action == 'store_false':
                        default = True
                        validator = pulsar.validate_bool
                setting = pulsar.Setting(name=setting_name,
                                    desc=attrs.pop('desc', name),
                                    type=type,
                                    flags=attrs.pop('flags', [def_flag]),
                                    action=action,
                                    default=default,
                                    validator=validator,
                                    nargs=nargs)
                settings[setting.name] = as_test_setting(setting)
        attrs['config'] = pulsar.Config(settings=settings)
        return super(TestPluginMeta, cls).__new__(cls, name, bases, attrs)
                
    
class TestPlugin(TestPluginMeta('TestPluginBase', (Plugin,), {'virtual': True})):
    '''Base class for :class:`result.Plugin` which can be added to a
:class:`TestSuite` to extend its functionalities. :ref:`Settings <settings>`
can be specified as class attributes and collected into the
:attr:`config` attribute.

If the class attribute :attr:`name` is not specified or its value validate as
``True``, an additional :ref:`setting <settings>` is added to the configuration.
In addition, a :class:`TestPlugin` can specify several additional
:ref:`settings <settings>` as class attributes. For example, the
:ref:`benchmark plugin <bench-plugin>` has an additional setting
for controlling the number of repetitions::

    class Bench(TestPlugin):
        desc = "Run benchmarks function flagged with __benchmark__ attribute"
        repeat = pulsar.Setting(type=int,
                            default=1,
                            validator=pulsar.validate_pos_int,
                            desc="Default number of repetition")


.. attribute:: name

    Class attribute used for adding the default plugin :ref:`setting <settings>`
    to the configuration container of the test suite application.
    If the attribute is not set, the class name in lower case is used.
    If set and validate as not ``True``, no new :ref:`setting <settings>`
    is added to the test suite configuration parameters. For example::
    
        class MyPlugin(TestPlugin):
            name = None
            
    won't add the default plugin :ref:`setting <settings>`.
    
.. attribute:: desc

    Class attribute used as the description of the :ref:`setting <settings>`.
    If :attr:`name` is disabled, this attribute is not relevant.
    
.. attribute:: config

    A :class:`pulsar.utils.config.Config` container created by th
    :class:`TestPlugin` metaclass. It collects the default setting, if
    available, and any additional :ref:`settings <setting>` specified
    as class attributes.
'''
    virtual = True
    
    def __new__(cls):
        o = super(TestPlugin, cls).__new__(cls)
        o.config = cls.config.copy()
        return o
        
    def configure(self, cfg):
        self.config = cfg
    


'''Tests the websocket middleware in pulsar.apps.ws.'''
from pulsar import send
from pulsar.apps import rpc, http
from pulsar.apps.test import unittest, dont_run_with_thread

from .manage import server


class TestWebChat(unittest.TestCase):
    app = None
    concurrency = 'thread'
    
    @classmethod
    def setUpClass(cls):
        s = server(bind='127.0.0.1:0', name=cls.__name__,
                   concurrency=cls.concurrency)
        cls.app = yield send('arbiter', 'run', s)
        cls.uri = 'http://%s:%s' % cls.app.address
        cls.ws = 'ws://%s:%s/message' % cls.app.address
        cls.rpc = rpc.JsonProxy('%s/rpc' % cls.uri)
        cls.http = http.HttpClient()
        
    @classmethod
    def tearDownClass(cls):
        if cls.app is not None:
            yield send('arbiter', 'kill_actor', cls.app.name)
            
    def test_rpc(self):
        '''Send a message to the rpc'''
        response = yield self.http.get(self.ws).on_finished
        self.assertEqual(response.handshake.status_code, 101)
        self.assertFalse(response.request_processed)
        result = yield self.rpc.message('Hi!')
        self.assertEqual(result, 'OK')
        
    def test_handshake(self):
        ws = yield self.http.get(self.ws).on_finished
        response = ws.handshake 
        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers['upgrade'], 'websocket')
        self.assertEqual(response.connection, None)
        self.assertTrue(ws.connection)
        


import sys
import time
import string
import threading
from inspect import istraceback

ispy3k = sys.version_info >= (3, 0)
ispy26 = sys.version_info < (2, 7)
ispy33 = sys.version_info >= (3, 3)

if ispy33:
    default_timer = time.monotonic
else:   #pragma    nocover
    default_timer = time.time
    
    
if ispy3k: # Python 3
    import pickle
    string_type = str
    ascii_letters = string.ascii_letters
    zip = zip
    map = map
    range = range
    chr = chr
    iteritems = lambda d : d.items()
    itervalues = lambda d : d.values()
    is_string = lambda s: isinstance(s, str)
    is_string_or_native_string = is_string

    def to_bytes(s, encoding=None, errors=None):
        '''Convert *s* into bytes'''
        if not isinstance(s, bytes):
            return ('%s' % s).encode(encoding or 'utf-8', errors or 'strict')
        elif not encoding or encoding == 'utf-8':
            return s
        else:
            d = s.decode('utf-8')
            return d.encode(encoding, errors or 'strict')

    def to_string(s, encoding=None, errors='strict'):
        """Inverse of to_bytes"""
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8', errors)
        else:
            return str(s)

    def native_str(s, encoding=None):
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8')
        else:
            return s

    def force_native_str(s, encoding=None):
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8')
        elif not isinstance(s, str):
            return str(s)
        else:
            return s

    def raise_error_trace(err, traceback):
        if istraceback(traceback):
            raise err.with_traceback(traceback)
        else:
            raise err
        
else:   # pragma : no cover
    from itertools import izip as zip, imap as map
    import cPickle as pickle
    from ._py2 import *
    string_type = unicode
    ascii_letters = string.letters
    range = xrange
    chr = unichr
    iteritems = lambda d : d.iteritems()
    itervalues = lambda d : d.itervalues()
    is_string = lambda s: isinstance(s, unicode)
    
    def to_bytes(s, encoding=None, errors='strict'):
        encoding = encoding or 'utf-8'
        if isinstance(s, bytes):
            if encoding != 'utf-8':
                return s.decode('utf-8', errors).encode(encoding, errors)
            else:
                return s
        else:
            return unicode(s).encode(encoding, errors)

    def to_string(s, encoding=None, errors='strict'):
        """Inverse of to_bytes"""
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8', errors)
        else:
            return unicode(s)

    def native_str(s, encoding=None):
        if isinstance(s, unicode):
            return s.encode(encoding or 'utf-8')
        else:
            return s

    def force_native_str(s, encoding=None):
        if isinstance(s, unicode):
            return s.encode(encoding or 'utf-8')
        elif not isinstance(s, str):
            return str(s)
        else:
            return s
        
        
################################################################################
###    PEP 3156
###    These classes will be eventually replaced by the standard lib

class EventLoop(object):
    '''This is just a signature'''
    
    
class EventLoopPolicy:
    """Abstract policy for accessing the event loop."""

    def get_event_loop(self):
        """XXX"""
        raise NotImplementedError

    def set_event_loop(self, event_loop):
        """XXX"""
        raise NotImplementedError

    def new_event_loop(self, **kwargs):
        """XXX"""
        raise NotImplementedError


# Event loop policy.  The policy itself is always global, even if the
# policy's rules say that there is an event loop per thread (or other
# notion of context).  The default policy is installed by the first
# call to get_event_loop_policy().
_event_loop_policy = None

def get_event_loop_policy():
    """XXX"""
    global _event_loop_policy
    return _event_loop_policy

def set_event_loop_policy(policy):
    """XXX"""
    global _event_loop_policy
    assert policy is None or isinstance(policy, EventLoopPolicy)
    _event_loop_policy = policy

def get_event_loop():
    """XXX"""
    return get_event_loop_policy().get_event_loop()

def set_event_loop(event_loop):
    """XXX"""
    get_event_loop_policy().set_event_loop(event_loop)

def new_event_loop(**kwargs):
    """XXX"""
    return get_event_loop_policy().new_event_loop(**kwargs)



'''
Plugin
~~~~~~~~~~~~~~~~~~~

.. autoclass:: Plugin
   :members:
   :member-order: bysource
   

Test Runner
~~~~~~~~~~~~~~~~~~~

.. autoclass:: TestRunner
   :members:
   :member-order: bysource

   
'''
import sys
import traceback
import unittest
import logging
from inspect import istraceback
from copy import deepcopy

from pulsar.utils.structures import AttributeDictionary

from .utils import TestFunction

__all__ = ['Plugin',
           'TestStream',
           'TestRunner',
           'TestResult',
           'Plugin',
           'LOGGER']


LOGGER = logging.getLogger('TestRunner')
STDOUT_LINE = '\nStdout:\n%s'
STDERR_LINE = '\nStderr:\n%s'


class Plugin(object):
    '''Interface for all classes which are part of of the :class:`TestRunner`,
including :class:`TestRunner` itself, :class:`TestResult`
and :class:`pulsar.apps.test.TestPlugin`.'''
    descriptions = None

    def configure(self, cfg):
        '''Called once just after construction of
a :class:`TestRunner` and therefore **before any test class is loaded**.
This is a chance to configure the :class:`Plugin` or global variables
which may affect the way tests are run.
If it returns something other than ``None`` (for example an abort message)
it will stop the configuration of all subsequent plugins and quit the test.

:parameter cfg: a :class:`pulsar.utils.config.Config`.
:return: ``None`` unless the tests runner must be stopped.
'''
        pass
    
    @property
    def name(self):
        return self.__class__.__name__.lower()

    def on_start(self):
        '''Called once by :class:`TestSuite` after all tests are loaded but
but before it starts running them.'''
        pass

    def on_end(self):
        '''Called once by :class:`TestSuite` just before it stops.'''
        pass

    def loadTestsFromTestCase(self, test_class):
        '''Called just before a ``test_class`` executes all its test
methods. This is run just before ``setUpClass`` method.'''
        pass
    
    def startTest(self, test):
        '''Called just before a ``test`` function is executed. This is run
just before ``_pre_setup`` method.'''
        pass

    def stopTest(self, test):
        '''Called just after a ``test`` function has finished. This is run just
after the ``_post_teardown`` method.'''
        pass
    
    def before_test_function_run(self, test, local):
        '''This function can be used by plugins to manipulate the *test*
behaviour in the process domain where the test run.'''
        return test
    
    def after_test_function_run(self, test, local, result, async):
        '''Given a test-function instance return a, possibly, modified test
instance. This function can be used by plugins to modify the behaviour of test
cases. By default it returns *test*.'''
        pass

    def addSuccess(self, test):
        pass

    def addFailure(self, test, err):
        pass

    def addError(self, test, err):
        pass

    def addSkip(self, test, reason):
        pass

    def printErrors(self):
        pass

    def printSummary(self, timeTaken):
        pass

    def import_module(self, mod, parent=None):
        return mod

    def getDescription(self, test):
        doc_first_line = test.shortDescription()
        teststr = '%s.%s' % (test.tag, test)
        if self.descriptions and doc_first_line:
            return '\n'.join((teststr, doc_first_line))
        else:
            return teststr


class TestResultProxy(Plugin):
    result = None
    stream = None

    @property
    def count(self):
        return self.result.count if self.result else 0

    @property
    def testsRun(self):
        return self.result.testsRun if self.result else 0


class TestStream(TestResultProxy):
    '''handle the writing of test results'''
    separator1 = '=' * 70
    separator2 = '-' * 70

    def __init__(self, stream, result, descriptions=True):
        self._handlers = {}
        self.stream = stream
        self.result = result
        self.descriptions = descriptions
        self.showAll = False
        self.dots = True

    def configure(self, cfg):
        verbosity = cfg.verbosity
        self.showAll = verbosity > 1
        self.dots = verbosity == 1

    def handler(self, name):
        return self._handlers.get(name, self.stream)

    def startTest(self, test):
        if self.showAll:
            self.head(test, 'Started')
            
    #def after_test_function_run(self, test, local, result, async):
    #    if async:
    #        self.startTest(test)

    def head(self, test, v):
        v = self.getDescription(test) + ' ... %s\n' % v
        self.stream.write(v)
        self.stream.flush()

    def addSuccess(self, test):
        if self.showAll:
            self.head(test, 'OK')
        elif self.dots:
            self.stream.write('.')
            self.stream.flush()

    def addError(self, test, err):
        if self.showAll:
            self.head(test, 'ERROR')
        elif self.dots:
            self.stream.write('E')
            self.stream.flush()

    def addFailure(self, test, err):
        if self.showAll:
            self.head(test, "FAIL")
        elif self.dots:
            self.stream.write('F')
            self.stream.flush()

    def addSkip(self, test, reason):
        if self.showAll:
            self.head(test,"skipped {0!r}".format(reason))
        elif self.dots:
            self.stream.write("s")
            self.stream.flush()

    def addExpectedFailure(self, test, err):
        if self.showAll:
            self.head(test, "expected failure")
        elif self.dots:
            self.stream.write("x")
            self.stream.flush()

    def addUnexpectedSuccess(self, test):
        if self.showAll:
            self.head(test, "unexpected success")
        elif self.dots:
            self.stream.write("u")
            self.stream.flush()

    def printErrors(self):
        if self.dots or self.showAll:
            self.stream.writeln()
        self.printErrorList('ERROR', self.result.errors)
        self.printErrorList('FAIL', self.result.failures)
        return True

    def printErrorList(self, flavour, errors):
        for test, err in errors:
            self.stream.writeln(self.separator1)
            self.stream.writeln("%s: %s" % (flavour,test))
            self.stream.writeln(self.separator2)
            self.stream.writeln("%s" % err)

    def printSummary(self, timeTaken):
        '''Write the summuray of tests results.'''
        stream = self.stream
        result = self.result
        self.printErrors()
        run = result.testsRun
        stream.writeln("Ran %d test%s in %.3fs" %
                            (run, run != 1 and "s" or "", timeTaken))
        stream.writeln()

        expectedFails = unexpectedSuccesses = skipped = 0
        results = map(len, (result.expectedFailures,
                            result.unexpectedSuccesses,
                            result.skipped))
        expectedFails, unexpectedSuccesses, skipped = results

        infos = []
        if not result.wasSuccessful():
            stream.write("FAILED")
            failed, errored = map(len, (result.failures, result.errors))
            if failed:
                infos.append("failures=%d" % failed)
            if errored:
                infos.append("errors=%d" % errored)
        else:
            stream.write("OK")
        if skipped:
            infos.append("skipped=%d" % skipped)
        if expectedFails:
            infos.append("expected failures=%d" % expectedFails)
        if unexpectedSuccesses:
            infos.append("unexpected successes=%d" % unexpectedSuccesses)
        if infos:
            stream.writeln(" (%s)" % (", ".join(infos),))
        else:
            stream.write("\n")

        return True


class TestResult(Plugin):
    '''A result of a test run.'''
    def __init__(self, descriptions=True):
        self.descriptions = descriptions
        self.testsRun = 0
        self._count = 0
        self.failures = []
        self.errors = []
        self.skipped = []
        self.expectedFailures = []
        self.unexpectedSuccesses = []

    @property
    def count(self):
        return self._count

    def startTest(self, test):
        self.testsRun += 1

    def addError(self, test, err):
        """Called when an error has occurred. 'err' is a tuple of values as
        returned by sys.exc_info().
        """
        self._add_error(test, err, self.errors)

    def addFailure(self, test, err):
        self._add_error(test, err, self.failures)

    def addSkip(self, test, reason):
        """Called when a test is skipped."""
        self.skipped.append((self.getDescription(test), reason))

    def addExpectedFailure(self, test, err):
        """Called when an expected failure/error occured."""
        self._add_error(test, err, self.expectedFailures)

    def addUnexpectedSuccess(self, test):
        """Called when a test was expected to fail, but succeed."""
        self.unexpectedSuccesses.append(self.getDescription(test))

    def _add_error(self, test, err, container):
        err = self._exc_info_to_string(err, test)
        test = self.getDescription(test)
        container.append((test, err))

    def add(self, result):
        self._count += 1
        self.testsRun += result.testsRun
        self.failures.extend(result.failures)
        self.errors.extend(result.errors)
        self.skipped.extend(result.skipped)
        self.expectedFailures.extend(result.expectedFailures)
        self.unexpectedSuccesses.extend(result.unexpectedSuccesses)

    def wasSuccessful(self):
        "Tells whether or not this result was a success"
        return len(self.failures) == len(self.errors) == 0

    def _exc_info_to_string(self, err, test):
        """Converts a sys.exc_info()-style tuple of values into a string."""
        exctype, value, tb = err
        # Skip test runner traceback levels
        if istraceback(tb):
            while tb and self._is_relevant_tb_level(tb):
                tb = tb.tb_next

            if exctype is test.failureException:
                # Skip assert*() traceback levels
                length = self._count_relevant_tb_levels(tb)
                msgLines = traceback.format_exception(exctype, value, tb, length)
            else:
                msgLines = traceback.format_exception(exctype, value, tb)
        else:
            msgLines = tb

        if not msgLines:
            if value:
                msgLines = [repr(value)]
            else:
                msgLines = ()

        return ''.join(msgLines)


    def _is_relevant_tb_level(self, tb):
        return '__unittest' in tb.tb_frame.f_globals

    def _count_relevant_tb_levels(self, tb):
        length = 0
        while tb and not self._is_relevant_tb_level(tb):
            length += 1
            tb = tb.tb_next
        return length


def testsafe(name, return_val=None):
    if not return_val:
        return_val = lambda c: None
    def _(self, *args):
        for p in self.plugins:
            try:
                c = getattr(p, name)(*args)
                if c:
                    return return_val(c)
            except Exception:
                LOGGER.critical('Unhadled error in %s.%s' % (p, name),
                                exc_info=True)
    return _                
            
    
class TestRunner(TestResultProxy):
    '''An asynchronous test runner'''
    def __init__(self, plugins, stream, writercls=None, descriptions=True,
                 logger=None):
        self.descriptions = descriptions
        self.plugins = []
        writercls = writercls or TestStream
        result = TestResult(descriptions=self.descriptions)
        stream = writercls(stream, result, descriptions=self.descriptions)
        for p in plugins:
            p = deepcopy(p)
            p.descriptions = self.descriptions
            p.result = result
            p.stream = stream
            self.plugins.append(p)
        self.plugins.append(result)
        self.plugins.append(stream)
        self.stream = stream
        self.result = result
        self.loader = unittest.TestLoader()

    configure = testsafe('configure', lambda c: c)        
    on_start = testsafe('on_start')
    on_end = testsafe('on_end')
    startTest = testsafe('startTest')    
    stopTest = testsafe('stopTest')
    addSuccess = testsafe('addSuccess')
    addFailure = testsafe('addFailure')
    addError = testsafe('addError')
    addSkip = testsafe('addSkip')
    printErrors = testsafe('printErrors')
    printSummary = testsafe('printSummary')

    def loadTestsFromTestCase(self, test_cls):
        '''Load all *test* functions for the test class *cls*.'''
        c = testsafe('loadTestsFromTestCase', lambda v: v)(self, test_cls)
        if c is None:
            return self.loader.loadTestsFromTestCase(test_cls)
        else:
            return c

    def add(self, result):
        self.result.add(result)

    def import_module(self, mod, parent=None):
        for p in self.plugins:
            mod = p.import_module(mod, parent)
            if not mod:
                return
        return mod

    def before_test_function_run(self, test):
        '''Called before the test run, in the test process domain.'''
        test.plugins = plugins = {}
        for p in self.plugins:
            local = AttributeDictionary()
            plugins[p.name] = local
            test = p.before_test_function_run(test, local) or test
        return test
    
    def after_test_function_run(self, test, result, async):
        '''Called before the test starts.'''
        for p in self.plugins:
            local = test.plugins.get(p.name)
            if local is not None:
                p.after_test_function_run(test, local, result, async)
        return result
            
    def run_test_function(self, test, func, timeout=None):
        '''Run function *func* which belong to *test*.
    
:parameter test: test instance or class
:parameter func: test function belonging to *test*
:parameter timeout: An optional timeout for asynchronous tests.
:return: an asynchronous result
'''
        if func is None:
            return func
        tfunc = getattr(func, 'testfunction', None)
        if tfunc is None:
            tfunc = TestFunction(func.__name__)
        return tfunc(test, timeout)


'''Operative system specific functions and classes.'''
from .runtime import Platform

platform = Platform()
seconds = platform.seconds

from .base import *

if platform.type == 'posix':
    from .posixsystem import *
elif platform.type == 'win':    #pragma nocover
    from .windowssystem import *


try:
    import psutil
except ImportError:    #pragma    nocover
    psutil = None
    
memory_symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
memory_size = dict(((s,1 << (i+1)*10) for i,s in enumerate(memory_symbols)))

def convert_bytes(b):
    '''Convert a number of bytes into a human readable memory usage, bytes,
kilo, mega, giga, tera, peta, exa, zetta, yotta'''
    if b is None:
        return '#NA'
    for s in reversed(memory_symbols):
        if b >= memory_size[s]:
            value = float(b) / memory_size[s]
            return '%.1f%sB' % (value, s)
    return "%sB" % b
    
def system_info(pid=None):
    '''Returns a dictionary of system information for the process with id *pid*.
It uses the psutil_ module for the purpose. If psutil_ is not available
it returns an empty dictionary.

.. _psutil: http://code.google.com/p/psutil/
'''
    if psutil is None:  #pragma    nocover
        return {}
    pid = pid or os.getpid()
    p = psutil.Process(pid)
    mem = p.get_memory_info()
    return {'memory': mem.rss,
            'memory_virtual': mem.vms,
            'cpu_percent': p.get_cpu_percent(),
            'nice': p.get_nice(),
            'num_threads': p.get_num_threads()}
    
    


import sys
import json
from collections import deque

from pulsar import AuthenticationError, CommandNotFound
from pulsar.utils.pep import default_timer

from .defer import Deferred
from .consts import *

__all__ = ['ActorProxyDeferred',
           'ActorProxy',
           'ActorProxyMonitor',
           'get_proxy',
           'command',
           'get_command']

global_commands_table = {}
        
        
def get_proxy(obj, safe=False):
    if isinstance(obj, ActorProxy):
        return obj
    elif hasattr(obj, 'proxy'):
        return get_proxy(obj.proxy)
    else:
        if safe:
            return None
        else:
            raise ValueError('"%s" is not an actor or actor proxy.' % obj)

def actorid(actor):
    return actor.identity() if hasattr(actor, 'identity') else actor

def get_command(name):
    '''Get the command function *name*'''
    command = global_commands_table.get(name.lower())
    if not command:
        raise CommandNotFound(name)
    return command
    
    
class command:
    '''Decorator for pulsar command functions.
    
:parameter ack: ``True`` if the command acknowledge the sender with a
    response. Usually is set to ``True`` (which is also the default value).
'''
    def __init__(self, ack=True):
        self.ack = ack
    
    def __call__(self, f):
        self.name = f.__name__.lower()
        def command_function(request, args, kwargs):
            return f(request, *args, **kwargs)
        command_function.ack = self.ack
        command_function.__name__ = self.name
        command_function.__doc__ = f.__doc__
        global_commands_table[self.name] = command_function
        return command_function
            

class ActorIdentity(object):
    
    def identity(self):
        return self.aid
    
    
class ActorProxyDeferred(Deferred, ActorIdentity):
    '''A :class:`Deferred` for an :class:`ActorProxy`. The callback will
be an :class:`ActorProxy` which will be received once the remote :class:`Actor`
is fully functional.

.. attribute:: aid

    The the remote :attr:`Actor` id
     
'''
    def __init__(self, aid, msg=None):
        super(ActorProxyDeferred,self).__init__()
        if isinstance(aid, ActorProxyMonitor):
            aid.callback = self
            self.aid = aid.aid
        else:
            self.aid = aid
            # Listen for the callbacks and errorbacks
            msg.add_both(self.callback)
    
    def __repr__(self):
        return '%s(%s)' % (self.__class__, self.aid)
    __str__ = __repr__
    
    
class ActorProxy(ActorIdentity):
    '''This is an important component in pulsar concurrent framework. An
instance of this class is as a proxy for a remote `underlying` 
:class:`Actor`. This is a lightweight class which delegates
function calls to the underlying remote object.

It is pickable and therefore can be send from actor to actor using pulsar
messaging. It exposes all the underlying :class:`command` which have been
implemented.

For example, lets say we have a proxy ``a``, to send a message to it::

    from pulsar import send
    
    send(a, 'echo', 'hello there!')
    
will send the command ``echo`` to actor ``a`` with
parameter ``"hello there!"``.
    
.. attribute:: aid

    Unique ID for the remote :class:`Actor`
    
.. attribute:: address

    the socket address of the underlying :attr:`Actor.mailbox`.
    
'''
    def __init__(self, impl):
        self.aid = impl.aid
        self.name = impl.name
        self.cfg = impl.cfg
        if hasattr(impl, 'address'):
            self.address = impl.address
    
    def __repr__(self):
        return '%s(%s)' % (self.name, self.aid)
    __str__ = __repr__
        
    @property
    def proxy(self):
        return self
    
    def __eq__(self, o):
        o = get_proxy(o,True)
        return o and self.aid == o.aid
    
    def __ne__(self, o):
        return not self.__eq__(o)
        

class ActorProxyMonitor(ActorProxy):
    '''A specialised :class:`ActorProxy` class which contains additional
information about the remote underlying :class:`pulsar.Actor`. Unlike the
:class:`pulsar.ActorProxy` class, instances of this class serialise
into their :attr:`proxy` attribute..

.. attribute:: impl

    The :class:`Concurrency` instance for the remote :class:`Actor`.
    
.. attribute:: info

    Dictionary of information regarding the remote :class:`Actor`
'''
    monitor = None
    def __init__(self, impl):
        self.impl = impl
        self.info = {}
        self.mailbox = None
        self.callback = None
        self.spawning_start = None
        self.stopping_start = None
        super(ActorProxyMonitor,self).__init__(impl)
        
    @property
    def notified(self):
        return self.info.get('last_notified')
    
    @property
    def pid(self):
        return self.impl.pid
    
    @property
    def proxy(self):
        return ActorProxy(self)
    
    def __reduce__(self):
        return self.proxy.__reduce__()
    
    def is_alive(self):
        '''True if underlying actor is alive'''
        return self.impl.is_alive()
        
    def terminate(self):
        '''Terminate life of underlying actor.'''
        self.impl.terminate()
    
    def join(self, timeout=None):
        '''Wait until the underlying actor terminates. If *timeout* is
provided, it raises an exception if the timeout is reached.'''
        self.impl.join(timeout=timeout)

    def start(self):
        '''Start the remote actor.'''
        self.spawning_start = default_timer()
        self.impl.start()
        
    def should_be_alive(self):
        return default_timer() - self.spawning_start > ACTOR_ACTION_TIMEOUT

    def should_terminate(self):
        if self.stopping_start is None:
            self.stopping_start = default_timer()
            return False
        else:
            return default_timer() - self.stopping_start > ACTOR_ACTION_TIMEOUT
    


import sys
import threading
import inspect
import ctypes
from multiprocessing import pool, dummy, current_process
from threading import Lock, ThreadError
from functools import partial

try:
    import queue
except ImportError: #pragma nocover
    import Queue as queue
ThreadQueue = queue.Queue
Empty = queue.Empty

from pulsar.utils.system import EpollInterface
from pulsar.utils.log import LocalMixin, local_property
from pulsar.utils.pep import set_event_loop

from .access import get_actor, set_actor, thread_local_data
from .eventloop import StopEventLoop, EventLoop
from .defer import maybe_async, is_async, log_failure


__all__ = ['Thread', 'IOQueue', 'ThreadPool', 'ThreadQueue', 'Empty']

def _async_raise(tid, exctype):
    """raises the exception, performs cleanup if needed"""
    if not inspect.isclass(exctype):
        raise TypeError("Only types can be raised (not instances)")
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("invalid thread id")
    elif res != 1:
        # """if it returns a number greater than one, you're in trouble, 
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
        raise SystemError("PyThreadState_SetAsyncExc failed")
    
    
class KillableThread(dummy.DummyProcess):
    '''A killable thread has the :meth:`terminate` method.
    
http://tomerfiliba.com/recipes/Thread2/
'''
    @property
    def pid(self):
        return current_process().pid
    
    # Internals
    def _get_my_tid(self):
        """determines this (self's) thread id"""
        if not self.isAlive():
            raise ThreadError("the thread is not active")
        # do we have it cached?
        if hasattr(self, "_thread_id"):
            return self._thread_id
        # no, look for it in the _active dict
        for tid, tobj in threading._active.items():
            if tobj is self:
                self._thread_id = tid
                return tid
        raise AssertionError("could not determine the thread's id")
    
    def terminate(self):
        '''First check if the thread has an event_loop attribute, if so
it invoke the stop method. Otherwise it raises SystemExit in the context of
the given thread, which should cause the thread to exit silently
(unless caught). Originally from http://tomerfiliba.com/recipes/Thread2/.'''
        event_loop = getattr(self, 'event_loop', None)
        if event_loop:
            self.event_loop.stop()
        else:
            _async_raise(self._get_my_tid(), SystemExit)
    
    
class Thread(KillableThread):
    '''This class should be used when creating threads in pulsar. It
makes sure the class:`Actor` controlling the thread is available.'''
    def __init__(self, *args, **kwargs):
        self.actor = get_actor()
        super(Thread, self).__init__(*args, **kwargs)
        self.name = '%s-%s' % (self.actor, self.name)
        
    def run(self):
        '''Modified run method which set the actor and the event_loop for
the running thread.'''
        actor = self.actor
        del self.actor
        set_actor(actor)
        set_event_loop(actor.event_loop)
        super(Thread, self).run()
        
    @property
    def event_loop(self):
        return thread_local_data('_request_loop', ct=self)


class _FdFactory:
    
    def __init__(self, fd, eventloop, read=None, **kwargs):
        self.fd = fd
        self.eventloop = eventloop
        self.handle_read = None
        self.add_reader(read)
    
    def add_connector(self, callback):
        pass
        
    def add_reader(self, callback):
        if not self.handle_read:
            self.handle_read = callback
        else:
            raise RuntimeError("Already reading!")
        
    def add_writer(self, callback):
        pass
        
    def remove_connector(self):
        pass
    
    def remove_writer(self):
        pass
    
    def remove_reader(self):
        '''Remove reader and return True if writing'''
        self.handle_read = None
    
    def __call__(self, request):
        return self.handle_read(request)


class IOQueue(EpollInterface, LocalMixin):
    '''Epoll like class for a IO based on queues rather than sockets.
The interface is the same as the python epoll_ implementation.

.. attribute:: queue

    The python ``Queue`` from where this poller get tasks at each
    iteration of the :class:`EventLoop`
    
.. attribute:: maxtasks

    Optional number of maximum tasks to process
    
.. attribute:: received

    Number of tasks received by this :class:`IOQueue`
    
.. attribute:: completed

    Number of tasks completed by this :class:`IOQueue`

.. _epoll: http://docs.python.org/library/select.html#epoll-objects'''
    fd_factory = _FdFactory
    
    def __init__(self, queue, maxtasks=None):
        assert maxtasks is None or (type(maxtasks) == int and maxtasks > 0)
        self.received = 0
        self.completed = 0
        self._wakeup = 0
        self._queue = queue
        self.maxtasks = maxtasks

    @property
    def cpubound(self):
        '''Required by the :class:`EventLoop` so that the event loop
install itself as a request loop rather than the IO event loop.'''
        return True
    
    def fileno(self):
        '''dummy file number'''
        return 0
    
    def register(self, fd, events=None):
        pass

    def modify(self, fd, events=None):
        pass

    def unregister(self, fd):
        pass

    def get(self, timeout=0.5):
        '''Wait for events. timeout in seconds (float)'''
        block = True
        with self.lock:
            if self._wakeup:
                block = False
                self._wakeup -= 1
        return self._queue.get(block=block, timeout=timeout)
    
    def poll(self, timeout=0.5):
        if self.maxtasks and self.received >= self.maxtasks:
            if self.completed < self.received:
                return ()
            else:
                raise StopEventLoop
        try:
            task = self.get(timeout=timeout)
        except (Empty, TypeError):
            return ()
        except (EOFError, IOError):
            raise StopEventLoop
        if task is None:    # got the sentinel, exit!
            raise StopEventLoop
        self.received += 1
        return ((self.fileno(), task),)
    
    def install_waker(self, loop):
        return self
        
    def wake(self):
        '''Waker implementation. This IOQueue is its own waker.'''
        with self.lock:
            self._wakeup += 1
            
    @local_property
    def lock(self):
        return Lock()


class PoolWorker(object):
    '''A pool worker which handles asynchronous results form
functions send to the task queue. Instances of this class are
initialised on a new pulsar :class:`Thread`.'''
    def __init__(self, inqueue, outqueue, initializer=None, initargs=(),
                 maxtasks=None):
        if hasattr(inqueue, '_writer'):
            inqueue._writer.close()
            outqueue._reader.close()
        if initializer is not None:
            initializer(*initargs)
        self.io_poller = IOQueue(inqueue, maxtasks)
        self.outqueue = outqueue
        # Create the event loop which get tasks from the task queue 
        event_loop = EventLoop(io=self.io_poller, poll_timeout=1)
        event_loop.add_reader(self.io_poller.fileno(), self._handle_request)
        event_loop.run_forever()
        
    def _handle_request(self, task):
        job, i, func, args, kwds = task
        try:
            result = func(*args, **kwds)
        except Exception:
            result = sys.exc_info()
        result = maybe_async(result)
        if is_async(result):
            result.add_both(partial(self._handle_result, job, i))
        else:
            self._handle_result(job, i, result)
            
    def _handle_result(self, job, i, result):
        log_failure(result)
        self.outqueue.put((job, i, (True, result)))
        self.io_poller.completed += 1


class ThreadPool(pool.ThreadPool):
    '''A modified :class:`multiprocessing.pool.ThreadPool` used by a pulsar
:class:`Actor` when it needs CPUbound workers to consume tasks
on a task queue. An actor can create a new :class:`ThreadPool` via
the :meth:`Actor.create_thread_pool` method.'''
    Process = Thread
    created = 0
    
    def _repopulate_pool(self):
        """Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        for i in range(self._processes - len(self._pool)):
            self.created = self.created + 1
            w = self.Process(target=PoolWorker,
                             args=(self._inqueue, self._outqueue,
                                   self._initializer,
                                   self._initargs, self._maxtasksperchild),
                             name='PoolThread%s' % self.created)
            self._pool.append(w)
            w.daemon = True
            w.start()
    
    def apply(self, *args, **kwargs):
        raise NotImplementedError('Use apply_async')


'''\
Pulsar Python Shell example::

    python pshell.py
'''
try:
    import pulsar
except ImportError: #pragma nocover
    import sys
    sys.path.append('../../')
from pulsar.apps.shell import PulsarShell 

if __name__ == '__main__':  #pragma nocover
    PulsarShell().start()



import os
import fcntl
import resource
import grp
import pwd
import signal
from multiprocessing import Pipe

from .base import *

__all__ = ['IOpoll',
           'Epoll',
           'close_on_exec',
           'Waker',
           'daemonize',
           'EXIT_SIGNALS',
           'get_uid',
           'get_gid',
           'get_maxfd']


import select
if hasattr(select, 'epoll'):
    IOpoll = select.epoll
else:   #pragma    nocover
    IOpoll = IOselect
    Epoll = IOselect

# standard signal quit
EXIT_SIGNALS = (signal.SIGINT, signal.SIGTERM, signal.SIGABRT, signal.SIGQUIT)
# Default maximum for the number of available file descriptors.
REDIRECT_TO = getattr(os, "devnull", "/dev/null")

def get_parent_id():
    return os.getppid()

def chown(path, uid, gid):
    try:
        os.chown(path, uid, gid)
    except OverflowError:
        os.chown(path, uid, -ctypes.c_int(-gid).value)
        
        
def close_on_exec(fd):
    if fd:
        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
    
def _set_non_blocking(fd):
    flags = fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK
    fcntl.fcntl(fd, fcntl.F_SETFL, flags)
    
def get_uid(user=None):
    if not user:
        return os.geteuid()
    elif user.isdigit() or isinstance(user, int):
        return int(user)
    else:
        return pwd.getpwnam(user).pw_uid
    
def get_gid(group=None):
    if not group:
        return os.getegid()
    elif group.isdigit() or isinstance(group, int):
        return int(group)
    else:
        return grp.getgrnam(group).gr_gid
    
def setpgrp():
    os.setpgrp()

def get_maxfd():
    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if (maxfd == resource.RLIM_INFINITY):
        maxfd = MAXFD
    return maxfd

def daemonize():    #pragma    nocover
    """Standard daemonization of a process. Code is based on the
ActiveState recipe at http://code.activestate.com/recipes/278731/"""
    if os.fork() == 0: 
        os.setsid()
        if os.fork() != 0:
            os.umask(0) 
        else:
            os._exit(0)
    else:
        os._exit(0)
    maxfd = get_maxfd()
    # Iterate through and close all file descriptors.
    for fd in range(0, maxfd):
        try:
            os.close(fd)
        except OSError:    # ERROR, fd wasn't open to begin with (ignored)
            pass
    os.open(REDIRECT_TO, os.O_RDWR)
    os.dup2(0, 1)
    os.dup2(0, 2)


class Waker(object):
    
    def __init__(self):
        r, w = Pipe(duplex=False)
        _set_non_blocking(r.fileno())
        _set_non_blocking(w.fileno())
        close_on_exec(r.fileno())
        close_on_exec(w.fileno())
        self._writer = w
        self._reader = r
        
    def __str__(self):
        return 'Pipe waker %s' % self.fileno()
        
    def fileno(self):
        return self._reader.fileno()
    
    def wake(self):
        try:
            self._writer.send(b'x')
        except IOError:
            pass

    def consume(self):
        r = self._reader
        try:
             while r.poll():
                 r.recv()
        except (IOError, EOFError):
            pass


if hasattr(select, 'epoll'):
    
    class Epoll(EpollInterface):
        
        def __init__(self, ep=None):
            self._epoll = ep or select.epoll()
        
        def close(self):
            self._epoll.close()
            
        def fileno(self):
            return self._epoll.fileno()
        
        def fromfd(self, fd):
            return self._epoll.fromfd(fd)
        
        def register(self, fd, events):
            return self._epoll.register(fd, events)
        
        def modify(self, fd, events):
            return self._epoll.modify(fd, events)
        
        def unregister(self, fd):
            return self._epoll.unregister(fd)
        
        def poll(self, timeout=-1):
            return self._epoll.poll(timeout=timeout)



'''
An application implements several :class:`Job`
classes which specify the way each :class:`pulsar.apps.tasks.task.Task` is run.
Each :class:`Job` class is a task-factory, therefore,
a :ref:`task <apps-taskqueue-task>` is always associated
with one :class:`Job`, which can be of two types:

* standard (:class:`Job`)
* periodic (:class:`PeriodicJob`), a generator of scheduled tasks.

.. _job-callable:

Implementing jobs
~~~~~~~~~~~~~~~~~~~~~~~

To define a job is simple, subclass from :class:`Job` and implement the
**job callable method**::

    from pulsar.apps import tasks

    class Addition(tasks.Job):

        def __call__(self, consumer, a, b):
            "Add two numbers"
            return a+b

The *consumer*, instance of :class:`pulsar.apps.tasks.task.TaskConsumer`,
is passed by the :ref:`Scheduler <apps-taskqueue-scheduler>` and should
always be the first positional argument in the callable function.
The remaining positional arguments and/or key-valued parameters are needed by
your job implementation.

A :ref:`job callable <job-callable>` can also return a
:ref:`coroutine <coroutine>` if it needs to perform asynchronous IO during its
execution::

    class Crawler(tasks.Job):

        def __call__(self, consumer, sample, size=10):
            response = yield http.request(...)
            content = response.content
            ...

Job class
~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: Job
   :members:
   :member-order: bysource

Periodic job
~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: PeriodicJob
   :members:
   :member-order: bysource

Job registry
~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: JobRegistry
   :members:
   :member-order: bysource

'''
from datetime import datetime, date, timedelta
from hashlib import sha1
import logging
import inspect

from pulsar.utils.pep import iteritems
from pulsar.utils.importer import import_modules
from pulsar.utils.security import gen_unique_id


__all__ = ['JobMetaClass', 'Job', 'PeriodicJob',
           'anchorDate', 'JobRegistry', 'create_task_id']


def create_task_id():
    return gen_unique_id()[:8]


class JobRegistry(dict):
    """Site registry for tasks."""

    def regular(self):
        """A generator of all regular jobs."""
        return self.filter_types("regular")

    def periodic(self):
        """A generator of all periodic jobs."""
        return self.filter_types("periodic")

    def register(self, job):
        """Register a job in the job registry.

        The task will be automatically instantiated if not already an
        instance.

        """
        if isinstance(job, JobMetaClass) and job.can_register:
            name = job.name
            self[name] = job()

    def filter_types(self, type):
        """Return a generator of all tasks of a specific type."""
        return ((job_name, job)
                    for job_name, job in iteritems(self)
                            if job.type == type)
        
    @classmethod
    def load(cls, paths):
        self = cls()
        for mod in import_modules(paths):
            for name in dir(mod):
                self.register(getattr(mod, name))
        return self


class JobMetaClass(type):

    def __new__(cls, name, bases, attrs):
        attrs['can_register'] = not attrs.pop('abstract', False)
        job_name = attrs.get("name", name).lower()
        log_prefix = attrs.get("log_prefix") or "pulsar"
        attrs["name"] = job_name
        logname = '%s.job.%s' % (log_prefix, name)
        attrs['logger'] = logging.getLogger(logname)
        return super(JobMetaClass, cls).__new__(cls, name, bases, attrs)



class Job(JobMetaClass('JobBase', (object,), {'abstract': True})):
    '''The Job class which is used in a distributed task queue.

.. attribute:: name

    The unique name which defines the Job and which can be used to retrieve
    it from the job registry.

.. attribute:: abstract

    If set to ``True`` (default is ``False``), the Job won't be registered
    with the :class:`JobRegistry`.

.. attribute:: type

    Type of Job, one of ``regular`` and ``periodic``.

.. attribute:: timeout

    An instance of a datetime.timedelta or ``None``. If set, it represents the
    time lag after which a task which did not start expires.

    Default: ``None``.

.. attribute:: can_overlap

    Boolean indicating if this job can generate overlapping tasks. It can
    also be a callable which accept the same input parameters as the job
    callable function.

    Default: ``True``.

.. attribute:: doc_syntax

    The doc string syntax.

    Default: ``markdown``

.. attribute:: logger

    an instance of a logger. Created at runtime.
'''
    abstract = True
    type = "regular"
    timeout = None
    expires = None
    doc_syntax = 'markdown'
    can_overlap = True

    def __call__(self, consumer, *args, **kwargs):
        '''The Jobs' task executed by the consumer. This function needs to be
implemented by subclasses.'''
        raise NotImplementedError("Jobs must define the run method.")

    def run_job(self, consumer, jobname, *args, **kwargs):
        '''Send a new task request to the task queue from within another
:class:`Task`. This allows tasks to act as tasks factories.

:parameter consumer: The :class:`TaskConsumer` handling the :class:`Task`.
:parameter jobname: The name of the :class:`Job` to run.
:parameter args: positional argument for the :ref:`job callable <job-callable>`.
:parameter kwargs: key-valued parameters for the
    :ref:`job callable <job-callable>`.
:rtype: A :class:`pulsar.ActorMessage` if **ack=True** was passed in the
    key-valued parameters, otherwise nothing.
'''
        kwargs['from_task'] = consumer.task_id
        return consumer.backend.run_job(jobname, args, kwargs)

    def make_task_id(self, args, kwargs):
        '''Get the task unique identifier.
This can be overridden by Job implementation.

:parameter args: tuple of positional arguments passed to the job callable.
:parameter kwargs: dictionary of key-valued parameters passed to
    the job callable.
:rtype: a native string.

Called by the :attr:`TaskQueue.scheduler` when creating a new task.
'''
        can_overlap = self.can_overlap
        if hasattr(can_overlap,'__call__'):
            can_overlap = can_overlap(*args, **kwargs)
        if can_overlap:
            return create_task_id()
        else:
            suffix = ''
            if args:
                suffix = ' args(%s)' % ', '.join((str(a) for a in args))
            if kwargs:
                suffix += ' kwargs(%s)' % ', '.join(('%s=%s' % (k, kwargs[k])\
                            for k in sorted(kwargs)))
            name = '%s%s' % (self.name, suffix)
            return sha1(name.encode('utf-8')).hexdigest()[:8]
        

class PeriodicJob(Job):
    '''A periodic :class:`Job` implementation.'''
    abstract  = True
    type      = "periodic"
    anchor    = None
    '''If specified it must be a :class:`datetime.datetime` instance.
It controls when the periodic Job is run.'''
    run_every = None
    '''Periodicity as a :class:`datetime.timedelta` instance.'''

    def __init__(self, run_every = None):
        self.run_every = run_every or self.run_every
        if self.run_every is None:
            raise NotImplementedError('Periodic Jobs must have\
 a run_every attribute set, "{0}" does not have one'.format(self.name))

    def is_due(self, last_run_at):
        """Returns tuple of two items ``(is_due, next_time_to_run)``,
where next time to run is in seconds. e.g.

* ``(True, 20)``, means the Job should be run now, and the next
    time to run is in 20 seconds.

* ``(False, 12)``, means the Job should be run in 12 seconds.

You can override this to decide the interval at runtime.
        """
        return self.run_every.is_due(last_run_at)



def anchorDate(hour=0, minute=0, second=0):
    '''Create an anchor date.'''
    td = date.today()
    return datetime(year=td.year, month=td.month, day=td.day,
                    hour=hour, minute=minute, second=second)





'''
This module implements the main classes for pulsar application framework. The
framework is built on top of pulsar asynchronous engine and allows to
implement servers with very little effort. The main classes here are:
:class:`Application` and :class:`MultiApp` which, has the name suggests, is
a factory of several :class:`Application` running on a single server.
The :class:`Configurator` is a mixin used as base class for both 
:class:`Application` and :class:`MultiApp`.

.. note::

    An instance of an :class:`Application` is pickable and therefore can be sent
    from actor to actor using the
    :ref:`actor message passing api <tutorials-messages>`.

Configurator
===============================
   
.. autoclass:: Configurator
   :members:
   :member-order: bysource
   
   
Application
===============================
   
.. autoclass:: Application
   :members:
   :member-order: bysource
   

Multi App
===============================
      
.. autoclass:: MultiApp
   :members:
   :member-order: bysource
   
   
Get application
=========================

.. autofunction:: get_application


Application Worker
===============================
   
.. autoclass:: Worker
   :members:
   :member-order: bysource


Application Monitor
===============================
   
.. autoclass:: ApplicationMonitor
   :members:
   :member-order: bysource
   

Backend
===================

.. autoclass:: Backend
   :members:
   :member-order: bysource
'''
import os
import sys
import logging
from hashlib import sha1
from inspect import getfile

import pulsar
from pulsar import Actor, Monitor, get_actor, EventHandler
from pulsar.utils.structures import OrderedDict
from pulsar.utils.pep import pickle
from pulsar.utils.sockets import parse_connection_string, get_connection_string
from pulsar.utils.log import LocalMixin
from pulsar.utils.importer import import_module
from pulsar.utils.security import gen_unique_id

__all__ = ['Application',
           'MultiApp',
           'Backend',
           'Worker',
           'ApplicationMonitor',
           'get_application']
    
    
def get_application(name):
    '''Fetch the :class:`Application` associated with *name* if available. This
function returns a :ref:`coroutine <coroutine>` and therefore it
must be used as an asynchronous component::

    ...
    app = yield pulsar.get_application('taskqueue')
    ...

'''
    actor = get_actor()
    if actor:
        if actor.is_arbiter():
            yield _get_app(actor, name)
        else:
            yield actor.send('arbiter', 'run', _get_app, name)
        
def _get_app(arbiter, name):
    monitor = arbiter.get_actor(name)
    if monitor:
        return monitor.params.app
        
def monitor_start(self):
    self.app.monitor_start(self)
    if not self.cfg.workers:
        self.app.worker_start(self)
    self.app.fire_event('start')
        
def monitor_stop(self):
    if not self.cfg.workers:
        self.app.worker_stop(self)
    self.app.monitor_stop(self)

def arbiter_config(cfg):
    cfg = cfg.copy()
    for setting in list(cfg.settings.values()):
        if setting.app: # Don't include application specific settings
            cfg.settings.pop(setting.name)
        elif not setting.is_global:
            if not getattr(setting, 'inherit', False):
                setting.set(setting.default)
    return cfg 


class Worker(Actor):
    '''An :class:`pulsar.Actor` for serving a pulsar :class:`Application`.'''
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__(*args, **kwargs)
        self.bind_event('start', self.app.worker_start)
        self.bind_event('stopping', self.app.worker_stopping)
        self.bind_event('stop', self.app.worker_stop)
        
    @property
    def app(self):
        '''The :class:`Application` served by this :class:`Worker`.'''
        return self.params.app
    
    def io_poller(self):
        '''Delegates the :meth:`pulsar.Actor.io_poller` method to the
:meth:`Application.io_poller` method of the :attr:`app` attribute.'''
        return self.app.io_poller(self)
    
    def info(self):
        data = super(Worker, self).info()
        return self.app.worker_info(self, data)
    

class ApplicationMonitor(Monitor):
    '''A :class:`pulsar.Monitor` for managing a pulsar :class:`Application`.'''
    actor_class = Worker
    
    def __init__(self, *args, **kwargs):
        super(ApplicationMonitor, self).__init__(*args, **kwargs)
        self.bind_event('start', monitor_start)
        self.bind_event('stop', monitor_stop)
        
    @property
    def app(self):
        '''The :class:`Application` served by this
:class:`ApplicationMonitor`.'''
        return self.params.app
        
    ############################################################################
    # Delegates Callbacks to the application
    def monitor_task(self):
        self.app.monitor_task(self)
        
    def actorparams(self):
        p = Monitor.actorparams(self)
        app = self.app
        if self.cfg.concurrency == 'thread':
            app = pickle.loads(pickle.dumps(app))
        p.update({'app': app,
                  'name': '{0}-worker'.format(app.name)})
        return self.app.actorparams(self, p)
    
    def info(self):
        data = super(ApplicationMonitor, self).info()
        if not self.cfg.workers:
            return self.app.worker_info(self, data)
        else:
            return self.app.monitor_info(self, data)
        


class AppEvents(EventHandler):
    ONE_TIME_EVENTS = ('ready', 'start', 'stop')
    

class Configurator(object):
    '''A mixin for configuring and loading the various necessities for any
given server running on :mod:`pulsar` concurrent framework.

:parameter name: to override the class :attr:`name` attribute.
:parameter description: to override the class :attr:`cfg.description` attribute.
:parameter epilog: to override the class :attr:`cfg.epilog` attribute.
:parameter version: Optional version of this application, it overrides the
    class :attr:`cfg.version` attribute.
:parameter argv: Optional list of command line parameters to parse, if
    not supplied the :attr:`sys.argv` list will be used. The parameter is
    only relevant if **parse_console** is ``True``.
:parameter parse_console: ``True`` (default) if the console parameters needs
    parsing.
:parameter script: Optional string which set the :attr:`script` attribute.
:parameter params: a dictionary of configuration parameters which overrides
    the defaults and the :attr:`cfg` class attribute. They will be overritten
    by a :ref:`config file <setting-config>` or command line
    arguments.   

.. attribute:: name
    
    The name is unique if this is an :class:`Application`. In this
    case it defines the application monitor name as well and can be access in
    the arbiter domain via the :func:`get_application` function.
    
.. attribute:: argv

    Optional list of command line parameters. If not available the
    :attr:`sys.argv` list will be used when parsing the console.
    
.. attribute:: cfg

    The :class:`pulsar.utils.config.Config` for this :class:`Configurator`.
    If set as class attribute it will be replaced during initialisation.

    Default: ``None``.

.. attribute:: parsed_console

    ``True`` if this application parsed the console before starting.
    
.. attribute:: script

    Full path of the script which starts the application or ``None``.
    Evaluated during initialization via the :meth:`python_path` method.
'''
    name = None
    cfg = None
    
    def __init__(self,
                 name=None,
                 description=None,
                 epilog=None,
                 version=None,
                 argv=None,
                 parse_console=True,
                 script=None,
                 cfg=None,
                 load_config=True,
                 **kwargs):
        cls = self.__class__
        self.name = name or cls.name or cls.__name__.lower()
        if load_config or not isinstance(cfg, pulsar.Config):
            cfg = cfg or {}
            cfg.update(kwargs)
            cfg = cls.create_config(cfg)
        self.cfg = cfg
        self.cfg.description = description or self.cfg.description
        self.cfg.epilog = epilog or self.cfg.epilog
        self.cfg.version = version or self.cfg.version
        self.argv = argv
        self.parsed_console = parse_console
        self.script = self.python_path(script)
    
    @property
    def version(self):
        '''Version of this :class:`Application`'''
        return self.cfg.version
    
    @property
    def root_dir(self):
        '''Root directory of this :class:`Configurator`. Evaluated from the
:attr:`script` attribute.'''
        if self.script:
            return os.path.dirname(self.script)
    
    def __repr__(self):
        return self.name

    def __str__(self):
        return self.__repr__()
    
    def python_path(self, script):
        '''Called during initialization to obtain the ``script`` name and
to add the :attr:`script` directory to the python path if not in the
path already.
If ``script`` does not evalueate to ``True`` it is evaluated from
the ``__main__`` import. Returns the real path of the python
script which runs the application.'''
        if not script:
            try:
                import __main__
            except ImportError:
                return
            script = getfile(__main__)
        script = os.path.realpath(script)
        path = os.path.dirname(script)
        if path not in sys.path:
            sys.path.insert(0, path)
        return script
    
    def on_config(self):
        '''Callback when configuration is loaded. This is a chance to do
 an application specific check before the concurrent machinery is put into
 place. If it returns ``False`` the application will abort.'''
        pass
    
    def load_config(self):
        '''Load the application configuration from a file and/or
from the command line. Called during application initialization.
The parameters overriding order is the following:

 * default parameters.
 * the *params* passed in the initialization.
 * the parameters in the optional configuration file
 * the parameters passed in the command line.
'''
        # get the actor if available and override default cfg values with those
        # from the actor
        actor = get_actor()
        if actor and actor.running:
            # actor available and running. unless argv is set, skip parsing
            if self.argv is None:
                self.parsed_console = False
            # copy global settings
            for setting in actor.cfg.settings.values():
                if setting.is_global:
                    self.cfg.set(setting.name, setting.get())
        for name in list(self.cfg.params):
            if name in self.cfg.settings:
                self.cfg.params.pop(name)
        # parse console args
        if self.parsed_console:
            parser = self.cfg.parser()
            opts = parser.parse_args(self.argv)
            config = getattr(opts, 'config', None)
            # set the config only if config is part of the settings
            if config is not None and self.cfg.config:
                self.cfg.config = config
        else:
            parser, opts = None, None
        #
        # Load up the config file if its found.
        self.cfg.params.update(self.cfg.import_from_module())
        #
        # Update the configuration with any command line settings.
        if opts:
            for k, v in opts.__dict__.items():
                if v is None:
                    continue
                self.cfg.set(k.lower(), v)
    
    @classmethod
    def create_config(cls, params, prefix=None, name=None):
        '''Create a new :class:`pulsar.utils.config.Config` container by
overriding default values with *params*.'''
        if cls.cfg:
            cfg = cls.cfg.copy(name=name, prefix=prefix)
            # update with latest settings
            cfg.update_settings()
            cfg.update(params)
        else:
            cfg = pulsar.Config(name=name, prefix=prefix, **params)
        return cfg
    
    
class Application(Configurator, pulsar.Pulsar):
    """An application interface for configuring and loading
the various necessities for any given server or distributed application running
on :mod:`pulsar` concurrent framework.
Applications can be of any sorts or forms and the library is shipped with
several battery included examples in the :mod:`pulsar.apps` framework module.

These are the most important facts about a pulsar :class:`Application`

* It derives from :class:`Configurator` so that it has all the functionalities
  to parse command line arguments and setup the :attr:`Configurator.cfg`.
* Instances must be pickable. If non-pickable data needs to be add on an
  :class:`Application` instance, it must be stored on the
  :attr:`Application.local` dictionary.
* When a new :class:`Application` is initialized,
  a new :class:`ApplicationMonitor` instance is added to the
  :class:`Arbiter`, ready to perform its duties.

:parameter callable: Initialise the :attr:`Application.callable` attribute.
:parameter load_config: If ``False`` the :meth:`Configurator.load_config`
    is not invoked. Default ``True``.
:parameter params: Passed to the :class:`Configurator` initialiser.

.. attribute:: callable

    Optional callable serving or configuring your application.
    If provided, the callable must be pickable, therefore it is either
    a function or a pickable object.

    Default ``None``
"""
    def __init__(self, callable=None, load_config=True, **params):
        '''Initialize a new :class:`Application` and add its
:class:`ApplicationMonitor` to the class:`pulsar.Arbiter`.

:parameter version: Optional version number of the application.

    Default: ``pulsar.__version__``

:parameter parse_console: flag for parsing console inputs. By default it parse
    only if the arbiter has not yet started.
'''
        super(Application, self).__init__(load_config=load_config, **params)
        self.local.events = AppEvents()
        self.callable = callable
        if load_config:
            self.load_config()
        self()
        
    def __call__(self, actor=None):
        if actor is None:
            actor = get_actor()
        monitor = None
        if actor and actor.is_arbiter():
            monitor = actor.get_actor(self.name)
        if monitor is None and (not actor or actor.is_arbiter()):
            self.cfg.on_start()
            self.configure_logging()
            self.fire_event('ready')
            arbiter = pulsar.arbiter(cfg=arbiter_config(self.cfg))
            if self.on_config() is not False:
                monitor = arbiter.add_monitor(ApplicationMonitor,
                                              self.name,
                                              app=self,
                                              cfg=self.cfg)
                self.cfg = monitor.cfg
        return self.event('start')
        
    @property
    def app(self):
        '''Returns ``self``. Implemented so that the :class:`ApplicationMonitor`
and the :class:`Application` have the same interface.'''
        return self

    @property
    def monitor(self):
        actor = get_actor()
        if actor:
            return actor.registered.get(self.name)

    def __setstate__(self, state):
        super(Application, self).__setstate__(state)
        self.local.events = AppEvents()
        
    def fire_event(self, name):
        return self.local.events.fire_event(name, self)
        
    def bind_event(self, name, callback):
        return self.local.events.bind_event(name, callback)
        
    def event(self, name):
        return self.local.events.event(name)

    def io_poller(self, worker):
        '''Called by :meth:`Worker.io_poller` method during the initialization
of the :class:`Worker` event loop. By default it does nothing so that
the event loop chooses the most suitable IO poller.'''
        return None

    def add_timeout(self, deadline, callback):
        self.arbiter.event_loop.add_timeout(deadline, callback)
    
    # WORKERS CALLBACKS
    def worker_start(self, worker):
        '''Called by the :class:`Worker` :meth:`pulsar.Actor.on_start`
:ref:`callback <actor-callbacks>` method.'''
        pass

    def worker_info(self, worker, data):
        return data
    
    def worker_stopping(self, worker):
        '''Invoked when the *worker* fires the *stopping* event.'''
        pass
    
    def worker_stop(self, worker):
        '''Called by the :class:`Worker` :meth:`pulsar.Actor.on_stop`
:ref:`callback <actor-callbacks>` method.'''
        pass

    # MONITOR CALLBACKS
    def actorparams(self, monitor, params):
        '''A chance to override the dictionary of parameters *params*
before a new :class:`Worker` is spawned. This method is invoked by
the :class:`ApplicationMonitor`. By default it does nothing.'''
        return params

    def monitor_start(self, monitor):
        '''Callback by :class:`ApplicationMonitor` when starting.
The application is now in the arbiter but has not yet started.'''
        pass

    def monitor_info(self, monitor, data):
        return data
    
    def monitor_stop(self, monitor):
        '''Callback by :class:`ApplicationMonitor` when stopping'''
        pass

    def monitor_task(self, monitor):
        '''Callback by :class:`ApplicationMonitor` at each event loop'''
        pass
    
    def start(self):
        '''Start the application if it wasn't already started.'''
        arbiter = pulsar.arbiter()
        if arbiter and self.name in arbiter.registered:
            arbiter.start()
        return self


class MultiApp(Configurator):
    '''A :class:`MultiApp` is a tool for creating several :class:`Application`
and starting them at once. It makes sure all :ref:`settings` for the
applications created are available in the command line.
The :meth:`build` is the only method which must be implemented by subclasses.
Check the :class:`examples.taskqueue.manage.server` class in the
:ref:`taskqueue example <tutorials-taskqueue>` for an example.
The :class:`MultiApp` derives from :class:`Configurator` and therefore
supports all its configuration utilities.

A minimal example usage::

    import pulsar
    
    class Server(pulsar.MultiApp):
        def build(self):
            yield self.new_app(TaskQueue)
            yield self.new_app(WSGIserver, prefix="rpc", callable=..., ...)
            yield self.new_app(WSGIserver, prefix="web", callable=..., ...)
'''
    _apps = None
    
    def build(self):
        '''Virtual method, must be implemented by subclasses and return an
iterable over results obtained from calls to the :meth:`new_app` method.'''
        raise NotImplementedError
    
    def apps(self):
        '''List of :class:`Application` for this :class:`MultiApp`.
The list is lazily loaded from the :meth:`build` method.'''
        if self._apps is None:
            # Add non-default settings values to the list of cfg params
            self.cfg.params.update(((s.name, s.value) for s in\
                       self.cfg.settings.values() if s.value != s.default))
            self.cfg.settings = {}
            self._apps =  []
            apps = OrderedDict(self.build())
            if not apps:
                return self._apps
            # Load the configuration (command line and config file)
            self.load_config()
            kwargs = self._get_app_params()
            for App, name, callable, cfg in self._iter_app(apps):
                settings = self.cfg.settings
                new_settings = {}
                for key in cfg:
                    setting = settings[key].copy()
                    if setting.orig_name and setting.orig_name != setting.name:
                        setting.name = setting.orig_name
                    new_settings[setting.name] = setting
                cfg.settings = new_settings
                kwargs.update({'name': name, 'cfg': cfg, 'callable': callable})
                self._apps.append(App(**kwargs))
        return self._apps
    
    def new_app(self, App, prefix=None, callable=None, **params):
        '''Invoke this method in the :meth:`build` method as many times
as the number of :class:`Application` required by this :class:`MultiApp`.

:param App: an :class:`Application` class.
:param prefix: The prefix to use for the application, the prefix is appended to
    the application :ref:`config parameters <settings>` and to the
    application name. Each call to this methjod must use a different value
    of for this parameter. It can be ``None``.
:param callable: optional callable (function of object) used during
    initialisation of *App* (the :class:`Application.callable`).
:param params: additional key-valued parameters used when creating
    an instance of *App*.
:return: a tuple used by the :meth:`apps` method.
'''
        params.update(self.cfg.params.copy())
        params.pop('name', None)
        prefix = prefix or ''
        if not prefix:
            name = self.name
            cfg = App.create_config(params, name=name)
        else:
            name = '%s_%s' % (prefix, self.name)
            cfg = App.create_config(params, prefix=prefix, name=name)
        for k in cfg.settings:
            if k not in self.cfg.settings:
                self.cfg.settings[k] = cfg.settings[k]
        return prefix, (App, name, callable, cfg)
            
    def __call__(self, actor=None):
        return pulsar.multi_async((app(actor) for app in self.apps()))
        
    def start(self):
        '''Use this method to start all applications at once.'''
        apps = self.apps()
        arbiter = pulsar.arbiter()
        if arbiter:
            arbiter.start()
    
    ##    INTERNALS
    def _iter_app(self, app_name_callables):
        main = app_name_callables.pop('', None)
        if not main:
            raise pulsar.ImproperlyConfigured('No main application in MultiApp')
        yield main
        for app in app_name_callables.values():
            yield app
    
    def _get_app_params(self):
        params = self.cfg.params.copy()
        for key, value in self.__dict__.items():
            if key.startswith('_'):
                continue
            elif key == 'parsed_console':
                key = 'parse_console'
            params[key] = value
        params['load_config'] = False
        return params
    
    
class Backend(LocalMixin):
    '''Base class for backends. It is the base class for
:ref:`publish/subscribe backend <apps-pubsub>` and for the
:ref:`taskqueue backend <apps-taskqueue>`.
A :class:`Backend` is never initialised directly, the
:meth:`Backend.make` classmethod is used instead.

.. attribute:: scheme

    The scheme for this backend (``local``, ``redis``, ``http``, ...).
    
.. attribute:: name

    Optional name of the :class:`Application` which created this
    :class:`Backend`. If not available the ``arbiter`` is used as owner
    of this :class:`Backend`.
    
.. attribute:: params

    Dictionary of additional parameters passed during initialisation.
    
.. attribute:: connection_string

    The connection string fro this backend.
    
.. attribute:: default_path

    A class attribute which represents the path from where to load backend
    implementations.
'''
    def __init__(self, scheme, connection_string, name=None, **params):
        self.scheme = scheme
        self.name = name or 'arbiter'
        self.connection_string = connection_string
        self.params = dict(self.setup(**params) or {})
        value = '%s\n%s' % (self.__class__.__name__, self.connection_string)
        self._id = sha1(value.encode('utf-8')).hexdigest()
        
    def setup(self, **params):
        return params
    
    @property
    def id(self):
        '''Identy for this backend, calculated from the class name and the
:attr:`connection_string`. Therefore if two backend instances from
the same class have the same connection string, than the :attr:`id`
is the same.'''
        return self._id
    
    @classmethod
    def make(cls, backend=None, name=None, **kwargs):
        '''Create a new :class:`Backend` from a *backend* connection string
which is of the form::

    scheme:://host?params

For example, the local backend is::

    local://
    
A redis backend could be::

    redis://127.0.0.1:6379?db=1&password=bla
    
:param backend: the connection string
:param kwargs: additional key-valued parameters used by the :class:`Backend`
    during initialisation.
    '''
        if isinstance(backend, cls):
            return backend
        backend = backend or 'local://'
        scheme, address, params = parse_connection_string(backend,
                                                          default_port=0)
        if scheme == 'local':
            if address[0] == '0.0.0.0':
                address = ('',)
            if name:
                params['name'] = name
        else:
            kwargs['name'] = name
        con_str = get_connection_string(scheme, address, params)
        params.update(kwargs)
        if 'timeout' in params:
            params['timeout'] = int(params['timeout'])
        path = cls.path_from_scheme(scheme)
        module = import_module(path)
        return getattr(module, cls.__name__)(scheme, con_str, **params)
    
    @classmethod
    def path_from_scheme(cls, scheme):
        '''A class method which returns the import dotted path for a given
``scheme``. Must be implemented by subclasses.'''
        raise NotImplementedError


from pulsar import send

from .jsonrpc import JSONRPC


__all__ = ['PulsarServerCommands']


class PulsarServerCommands(JSONRPC):
    '''Some useful commands to add to your :class:`JSONRPC` handler to get you
started. It exposes the following functions:'''    
    def rpc_ping(self, request):
        '''Ping the server.'''
        return 'pong'
    
    def rpc_echo(self, request, message=''):
        '''Echo the server.'''
        return message
    
    def rpc_server_info(self, request):
        '''Return a dictionary of information regarding the server and workers.
It invokes the :meth:`extra_server_info` for adding custom information.'''
        info = yield send('arbiter', 'info')
        yield self.extra_server_info(request, info)
    
    def rpc_functions_list(self, request):
        '''List of (method name, method document) pair of all method exposed by
this :class:`JSONRPC` handler.'''
        return list(self.listFunctions())
    
    def rpc_documentation(self, request):
        '''Documentation in restructured text.'''
        return self.docs()
    
    def rpc_kill_actor(self, request, aid):
        '''Kill the actor with id equal to *aid*.'''
        return send('arbiter', 'kill_actor', aid)
        
    def extra_server_info(self, request, info):
        '''Not an ``rpc`` method, but an internal method which adds additional
information to the info dictionary. Used by the :meth:`rpc_server_info`
method.'''
        return info
    



import sys
import json
import time
import re
import textwrap
import logging
import traceback
from datetime import datetime, timedelta
from inspect import istraceback
from email.utils import formatdate


import pulsar
from pulsar.utils.structures import MultiValueDict
from pulsar import is_failure, HttpException, maybe_async, is_async, get_actor
from pulsar.utils.html import escape
from pulsar.utils.pep import to_string
from pulsar.utils.httpurl import responses, has_empty_content, ispy3k,\
                                 REDIRECT_CODES, iteritems, ENCODE_URL_METHODS,\
                                 parse_qsl, HTTPError
                                 
from .structures import Accept, MIMEAccept, CharsetAccept, LanguageAccept
from .content import Html

__all__ = ['handle_wsgi_error',
           'wsgi_error_msg',
           'render_error_debug',
           'async_wsgi',
           'wsgi_request',
           'set_wsgi_request_class',
           'HOP_HEADERS']

HOP_HEADERS = frozenset((
        'connection',
        'keep-alive',
        'proxy-authenticate',
        'proxy-authorization',
        'te',
        'trailers',
        'transfer-encoding',
        'upgrade',
        'server',
        'date',
    ))

LOGGER = logging.getLogger('pulsar.wsgi')

_RequestClass = None

def wsgi_request(environ, app_handler=None, urlargs=None):
    global _RequestClass
    return _RequestClass(environ, app_handler=app_handler, urlargs=urlargs)

def set_wsgi_request_class(RequestClass):
    global _RequestClass
    _RequestClass = RequestClass
    
    
def wsgi_iterator(gen, encoding):
    for data in gen:
        if isinstance(data, bytes):
            yield data
        else:
            yield data.encode(encoding)

def async_wsgi(request, result, callback):
    result = maybe_async(result)
    while is_async(result):
        yield b''
        result = maybe_async(result)
    for b in callback(request, result):
        yield b

def cookie_date(epoch_seconds=None):
    """Formats the time to ensure compatibility with Netscape's cookie
    standard.

    Accepts a floating point number expressed in seconds since the epoch in, a
    datetime object or a timetuple.  All times in UTC.  The :func:`parse_date`
    function can be used to parse such a date.

    Outputs a string in the format ``Wdy, DD-Mon-YYYY HH:MM:SS GMT``.

    :param expires: If provided that date is used, otherwise the current.
    """
    rfcdate = formatdate(epoch_seconds)
    return '%s-%s-%s GMT' % (rfcdate[:7], rfcdate[8:11], rfcdate[12:25])

def set_cookie(cookies, key, value='', max_age=None, expires=None, path='/',
                domain=None, secure=False, httponly=False):
    '''Set a cookie key into the cookies dictionary *cookies*.'''
    cookies[key] = value
    if expires is not None:
        if isinstance(expires, datetime):
            delta = expires - expires.utcnow()
            # Add one second so the date matches exactly (a fraction of
            # time gets lost between converting to a timedelta and
            # then the date string).
            delta = delta + timedelta(seconds=1)
            # Just set max_age - the max_age logic will set expires.
            expires = None
            max_age = max(0, delta.days * 86400 + delta.seconds)
        else:
            cookies[key]['expires'] = expires
    if max_age is not None:
        cookies[key]['max-age'] = max_age
        # IE requires expires, so set it if hasn't been already.
        if not expires:
            cookies[key]['expires'] = cookie_date(time.time() + max_age)
    if path is not None:
        cookies[key]['path'] = path
    if domain is not None:
        cookies[key]['domain'] = domain
    if secure:
        cookies[key]['secure'] = True
    if httponly:
        cookies[key]['httponly'] = True

_accept_re = re.compile(r'([^\s;,]+)(?:[^,]*?;\s*q=(\d*(?:\.\d+)?))?')

def parse_accept_header(value, cls=None):
    """Parses an HTTP Accept-* header.  This does not implement a complete
    valid algorithm but one that supports at least value and quality
    extraction.

    Returns a new :class:`Accept` object (basically a list of ``(value, quality)``
    tuples sorted by the quality with some additional accessor methods).

    The second parameter can be a subclass of :class:`Accept` that is created
    with the parsed values and returned.

    :param value: the accept header string to be parsed.
    :param cls: the wrapper class for the return value (can be
                         :class:`Accept` or a subclass thereof)
    :return: an instance of `cls`.
    """
    if cls is None:
        cls = Accept
    if not value:
        return cls(None)
    result = []
    for match in _accept_re.finditer(value):
        quality = match.group(2)
        if not quality:
            quality = 1
        else:
            quality = max(min(float(quality), 1), 0)
        result.append((match.group(1), quality))
    return cls(result)

def parse_cache_control_header(value, on_update=None, cls=None):
    """Parse a cache control header.  The RFC differs between response and
    request cache control, this method does not.  It's your responsibility
    to not use the wrong control statements.

    .. versionadded:: 0.5
       The `cls` was added.  If not specified an immutable
       :class:`~werkzeug.datastructures.RequestCacheControl` is returned.

    :param value: a cache control header to be parsed.
    :param on_update: an optional callable that is called every time a value
                      on the :class:`~werkzeug.datastructures.CacheControl`
                      object is changed.
    :param cls: the class for the returned object.  By default
                :class:`~werkzeug.datastructures.RequestCacheControl` is used.
    :return: a `cls` object.
    """
    if cls is None:
        cls = RequestCacheControl
    if not value:
        return cls(None, on_update)
    return cls(parse_dict_header(value), on_update)


def _gen_query(query_string, encoding):
    # keep_blank_values=True
    for key, value in parse_qsl((query_string or ''), True):
        yield to_string(key, encoding, errors='replace'),\
              to_string(value, encoding, errors='replace')

def query_dict(query_string, encoding='utf-8'):
    if query_string:
        return MultiValueDict(_gen_query(query_string, encoding))
    else:
        return MultiValueDict()
    
    
error_messages = {
    500: 'An exception has occurred while evaluating your request.',
    404: 'Cannot find what you are looking for.'
}

def wsgi_error_msg(response, msg):
    if response.content_type == 'application/json':
        return json.dumps({'status': response.status_code,
                           'message': msg})
    else:
        return msg
    
class dump_environ(object):
    __slots__ = ('environ',)
    
    def __init__(self, environ):
        self.environ = environ
        
    def __str__(self):
        env = iteritems(self.environ)
        return '\n%s\n' % '\n'.join(('%s = %s' % (k, v) for k, v in env))
    
    
def handle_wsgi_error(environ, trace=None):
    '''The default handler for errors while serving an Http requests.

:parameter environ: The WSGI environment.
:parameter trace: the error traceback. If not avaiable it will be obtained from
    ``sys.exc_info()``.
:parameter content_type: Optional content type.
:parameter encoding: Optional charset.
:return: a :class:`WsgiResponse`
'''
    if not trace:
        trace = sys.exc_info()
    error = trace[1]
    request = wsgi_request(environ)
    response = request.response
    content_type = response.content_type
    content = None
    if isinstance(error, HTTPError):
        response.status_code = error.code or 500
    else:
        response.status_code = getattr(error, 'status', 500)
        response.headers.update(getattr(error, 'headers', None) or ())
    path = ' @ path "%s"' % environ.get('PATH_INFO','/')
    e = dump_environ(environ)
    if response.status_code == 500:
        LOGGER.critical('Unhandled exception during WSGI response %s.%s',
                        path, e, exc_info=trace)
    else:
        LOGGER.warning('WSGI %s status code %s.', response.status_code, path)
        LOGGER.debug('%s', e, exc_info=trace)
    if has_empty_content(response.status_code) or\
            response.status_code in REDIRECT_CODES:
        response.content_type = None
    else:
        request.cache.pop('html_document', None)
        renderer = environ.get('error.handler') or render_error
        try:
            content = renderer(request, trace)
            if is_failure(content):
                content.log()
                content = None
        except Exception:
            LOGGER.critical('Error while rendering error', exc_info=True)
            content = None
    response.content = content or ()
    return response
           
def render_error(request, exc_info): 
    debug = get_actor().cfg.debug
    response = request.response
    if response.content_type == 'text/html':
        request.html_document.head.title = response.status
    if debug:
        msg = render_error_debug(request, exc_info)
    else:
        msg = error_messages.get(response.status_code) or ''
        if response.content_type == 'text/html':
            msg = textwrap.dedent("""
                <h1>{0[reason]}</h1>
                {0[msg]}
                <h3>{0[version]}</h3>
            """).format({"reason": response.status, "msg": msg,
                         "version": pulsar.SERVER_SOFTWARE})
    #
    if response.content_type == 'text/html':
        request.html_document.body.append(msg)
        return request.html_document.render(request)
    else:
        return wsgi_error_msg(response, msg)

def render_error_debug(request, exc_info):
    '''Render the traceback into the content type in *response*.'''
    if exc_info:
        response = request.response
        trace = exc_info[2]
        if istraceback(trace):
            trace = traceback.format_exception(*exc_info)
        is_html = response.content_type == 'text/html'
        if is_html:
            error = Html('div', cn='section traceback error')
        else:
            error = []
        for traces in trace:
            counter = 0
            for trace in traces.split('\n'):
                if trace.startswith('  '):
                    counter += 1
                    trace = trace[2:]
                if not trace:
                    continue
                if is_html:
                    trace = Html('p', escape(trace))
                    if counter:
                        trace.css({'margin-left':'%spx' % (20*counter)})
                error.append(trace)
        if not is_html:
            error = '\n'.join(error)
        return error


'''This example is a simple WSGI_ script which displays
the "Hello World!" message. To run the script simply type::

    python manage.py
    
To see all options available type::

    python manage.py -h

.. autofunction:: hello

.. _WSGI: http://www.python.org/dev/peps/pep-3333/ 
'''
try:
    import pulsar
except ImportError: #pragma nocover
    import sys
    sys.path.append('../../')
from pulsar.apps import wsgi


def hello(environ, start_response):
    '''The WSGI_ application handler which returns an iterable
over the "Hello World!" message.'''
    data = b'Hello World!\n'
    status = '200 OK'
    response_headers = [
        ('Content-type','text/plain'),
        ('Content-Length', str(len(data)))
    ]
    start_response(status, response_headers)
    return iter([data])


def server(description=None, **kwargs):
    description = description or 'Pulsar Hello World Application'
    return wsgi.WSGIServer(callable=hello,
                           description=description,
                           **kwargs)
    

if __name__ == '__main__':  #pragma nocover
    server().start()



'''Asynchronous application for serving requests
on a socket. This is the base class of :class:`pulsar.apps.wsgi.WSGIServer`.

A simple echo server ``script.py`` can be implemented as follow::

    import pulsar
    from pulsar.apps.socket import SocketServer
    
    class EchoProtocol:
        "The simplest protocol possible."
        def decode(self, data):
            return bytes(data), bytearray()
        
        def encode(self, data):
            return data
    
    def echoserver(actor, socket, **kwargs):
        kwargs['protocol_factory'] = EchoProtocol
        return pulsar.AsyncSocketServer(actor, socket, **kwargs)
        
    if __name__ == '__main__':
        SocketServer(socket_server_factory=echoserver).start()
    
    
The main parameter for :class:`SocketServer` is the ``socket_server_factory``
callable which build a :class:`pulsar.AsyncSocketServer` for each
:class:`pulsar.Worker` serving the application.

.. _socket-protocol:

Protocol
==============

The protocol class needs to implement two methods, the ``decode`` for servers
and the ``encode`` for clients. Both methods accept as input one parameter which
can be ``bytes`` or a ``bytesarray``.

The ``decode`` method must return a **two elements tuple** of the form:

 * ``None``, ``bytesarray``: when more data is needed to form a message.
 * ``message``, ``bytesarray``: where *message* is a correctly parsed message
   and *bytesarray* contains data for the next message. 

The ``encode`` method must return ``bytes`` which will be sent to the server.

.. _socket-server-settings:

Socket Server Settings
==============================
All standard :ref:`application settings <settings>` can be applied to a
:class:`SocketServer`. In addition, the following are
specific to sockets and can be used to fine tune your application:

bind
------
The :class:`SocketServer` application introduces the additional pulsar settings
:ref:`bind <setting-bind>` which is used to specify the address
to bind the server to::

    python script.py --bind 127.0.0.1:8070

backlog
---------   
To control the concurrency of the server you can use the
:ref:`backlog <setting-backlog>` settings. For example::

    python script.py --backlog 1000
    
will serve a maximum of 1000 clients per :class:`pulsar.Worker` concurrently.

keep_alive
---------------
To control how long a :class:`pulsar.AsyncConnection` is kept alive after the
last read from the remote client, one can use the
:ref:`keepalive <setting-keep_alive>` setting.

    python script.py --keep-alive 10
    
will close a client connection which has been idle for 10 seconds.
 


Concurrency
==================

When running a :class:`SocketServer` in threading mode::

    python script.py --concurrency thread
 
the number of :class:`pulsar.Worker` serving the application is set to 0 so
that the application is actually served by the :class:`Arbiter`
eventloop (we refer this to a single process server).
This configuration is used when debugging, testing, benchmarking or small
load servers.

In addition, a :class:`SocketServer` in multi-process mode is only available for:
    
 * Posix systems
 * Windows running python 3.2 or above.
'''
import pulsar


class Bind(pulsar.Setting):
    section = "Socket Servers"
    app = 'socket'
    name = "bind"
    flags = ["-b", "--bind"]
    meta = "ADDRESS"
    default = "127.0.0.1:{0}".format(pulsar.DEFAULT_PORT)
    can_prefix = True
    desc = """\
        The socket to bind.
        
        A string of the form: ``HOST``, ``HOST:PORT``, ``unix:PATH``.
        An IP is a valid HOST.
        """
        
class KeepAlive(pulsar.Setting):
    section = "Socket Servers"
    app = 'socket'
    name = "keep_alive"
    flags = ["--keep-alive"]
    validator = pulsar.validate_pos_int
    type = int
    default = 15
    can_prefix = True
    desc = """\
        The number of seconds to keep an idle client connection
        open."""


class SocketServer(pulsar.Application):
    '''This application bind a socket to a given address and listen for
requests. The request handler is constructued from the
:attr:`handler` parameter/attribute.
    
.. attribute:: address

    The socket address, available once the application has started.
    '''
    name = 'socket'
    address = None
    cfg = pulsar.Config(apps=['socket'])
    
    def protocol_consumer(self):
        '''Returns the callable application handler which is stored in
:attr:`Worker.app_handler`, used by :class:`Worker` to carry out its task.
By default it returns the :attr:`Application.callable`.'''
        return self.callable
    
    def monitor_start(self, monitor):
        # if the platform does not support multiprocessing sockets set
        # the number of workers to 0.
        cfg = self.cfg
        if not pulsar.platform.has_multiProcessSocket\
            or cfg.concurrency == 'thread':
            cfg.set('workers', 0)
        # Open the socket and bind to address
        address = self.cfg.address
        if address:
            sock = pulsar.create_socket(address, bindto=True,
                                        backlog=self.cfg.backlog)
        else:
            raise pulsar.ImproperlyConfigured('Could not open a socket. '
                                              'No address to bind to')
        self.logger.info('Listening on %s', sock)
        monitor.params.sock = sock
        self.address = sock.address
    
    def worker_start(self, worker):
        # Start the worker by starting the socket server
        worker.servers[self.name] = self.create_server(worker)
    
    def on_info(self, worker, data):
        server = worker.socket_server
        data['socket'] = {'listen_on': server.address,
                          'read_timeout': server.timeout,
                          'active_connections': server.active_connections,
                          'received_connections': server.received}
        return data

    def create_server(self, worker):
        '''Create the Server Protocol which will listen for requests. It
uses the :meth:`handler` as its response protocol.'''
        cfg = self.cfg
        server = worker.event_loop.create_server(
                                      sock=worker.params.sock,
                                      consumer_factory=self.protocol_consumer(),
                                      max_connections=cfg.max_requests,
                                      timeout=cfg.keep_alive)
        server.bind_event('connection_made', cfg.connection_made)
        server.bind_event('pre_request', cfg.pre_request)
        server.bind_event('post_request', cfg.post_request)
        server.bind_event('connection_lost', cfg.connection_lost)
        return server
        


'''This example creates two :ref:`pulsar applications <apps-framework>`
performing different duties. The first application is a distributed
a :ref:`task queue <apps-taskqueue>` for processing tasks implemented
in the :mod:`examples.taskqueue.simpletasks` module.
The second application is a :ref:`WSGI server <apps-wsgi>` which
exposes the task queue functionalities via a :ref:`JSON-RPC api <apps-rpc>`.

To run the server type::

    python manage.py
    
Open a new shell and launch python and type::

    >>> from pulsar.apps import rpc
    >>> p = rpc.JsonProxy('http://localhost:8060')
    >>> p.ping()
    'pong'
    >>> p.calc.add(3,4)
    7.0
    >>>

Implementation
====================

.. autoclass:: RpcRoot
   :members:
   :member-order: bysource
   
.. autoclass:: server
   :members:
   :member-order: bysource    
'''
try:
    import pulsar
except ImportError:
    import sys
    sys.path.append('../../')
    import pulsar
from pulsar.utils.path import Path
from pulsar.apps import rpc, tasks, wsgi

Path(pulsar.__file__).add2python('stdnet', up=2, down=['python-stdnet'],
                                 must_exist=False)

TASK_PATHS = ['sampletasks.*']


class RpcRoot(rpc.PulsarServerCommands, tasks.TaskQueueRpcMixin):
    '''The :class:`pulsar.apps.rpc.JSONRPC` handler which communicates
with the task queue.'''
    
    def rpc_runpycode(self, request, code=None, **params):
        return self.task_run(request, 'runpycode', code=code, **params)
        

class Rpc(wsgi.LazyWsgi):
    
    def __init__(self, tqname):
        self.tqname = tqname
        
    def setup(self):
        # only post allowed by the JSON RPC handler
        request = [wsgi.Router('/', post=RpcRoot(self.tqname))]
        response = [wsgi.GZipMiddleware(200)]
        return wsgi.WsgiHandler(middleware=request,
                                response_middleware=response)
    
    
class server(pulsar.MultiApp):
    '''Build a multi-app consisting on a taskqueue and a JSON-RPC server.
This class shows how to
use :class:`pulsar.apps.MultiApp` utility for starting several
:ref:`pulsar applications <apps-framework>` at once.'''
    cfg = pulsar.Config('Taskqueue with JSON-RPC API example')
    def build(self):
        yield self.new_app(tasks.TaskQueue, task_paths=TASK_PATHS)
        yield self.new_app(wsgi.WSGIServer, prefix='rpc',
                           callable=Rpc(self.name))
    

if __name__ == '__main__':
    server('taskqueue').start()




from .consts import *
from .access import *
from .defer import *
from .proxy import *
from .transports import *
from .eventloop import *
from .threads import *
from .actor import *
from .arbiter import *
from .monitor import *
from .concurrency import *
from . import commands


'''\
A very Simple Web-Socket example.
To run the server type::

    python manage.py
    
and open a web browser at http://localhost:8060    
'''
import os
import sys
import json
from random import random
import time
try:
    import pulsar
except ImportError: #pragma nocover
    sys.path.append('../../')
    import pulsar
from pulsar.apps import ws, wsgi
from pulsar.utils.httpurl import range


class Graph(ws.WS):
    
    def on_message(self, request, msg):
        return json.dumps([(i,random()) for i in range(100)])

    
class Echo(ws.WS):
    
    def on_message(self, request, msg):
        return msg


class Site(wsgi.LazyWsgi):
    
    def setup(self):
        return wsgi.WsgiHandler([wsgi.Router('/', get=self.home),
                                 ws.WebSocket('/data', Graph()),
                                 ws.WebSocket('/echo', Echo())])
    
    def home(self, request):
        data = open(os.path.join(os.path.dirname(__file__), 
                     'websocket.html')).read()
        data = data % request.environ
        request.response.content_type = 'text/html'
        request.response.content = data
        return request.response.start()


def server(**kwargs):
    return wsgi.WSGIServer(callable=Site(), **kwargs)


if __name__ == '__main__':  #pragma nocover
    server().start()



#    ORIGINAL FILE FROM TWISTED twisted.python.runtime
#    Modified and adapted for pulsar
#
# -*- test-case-name: twisted.python.test.test_runtime -*-
# Copyright (c) 2001-2008 Twisted Matrix Laboratories.
# See LICENSE for details.


# System imports
import os
import sys
import time
import imp
import socket


knownPlatforms = {
    'nt': 'win',
    'ce': 'win',
    'posix': 'posix',
    'java': 'java',
    'org.python.modules.os': 'java',
    }

_timeFunctions = {
    #'win': time.clock,
    'win': time.time,
    }

class Platform(object):
    """Gives us information about the platform we're running on"""

    name = os.name
    type = knownPlatforms.get(os.name)
    seconds = staticmethod(_timeFunctions.get(type, time.time))

    def __str__(self):
        return '{0} - {1}'.format(self.type, self.name)
    
    def __repr__(self):
        return '{0}: {1}'.format(self.__class__.__name__, self)
    
    def isKnown(self):
        """Do we know about this platform?"""
        return self.type != None

    def getType(self):
        """Return ``posix``, ``win`` or ``java``"""
        return self.type

    @property
    def is_posix(self):
        return self.type == 'posix'
    
    @property
    def isMacOSX(self):
        """Return if we are runnng on Mac OS X."""
        return sys.platform == "darwin"

    @property
    def is_winNT(self):
        """Are we running in Windows NT?"""
        if self.getType() == 'win':
            import _winreg
            try:
                k=_winreg.OpenKeyEx(_winreg.HKEY_LOCAL_MACHINE,
                                    r'Software\Microsoft\Windows NT\CurrentVersion')
                _winreg.QueryValueEx(k, 'SystemRoot')
                return True
            except WindowsError:
                return False
        # not windows NT
        return False

    @property
    def is_windows(self):
        return self.getType() == 'win'

    @property
    def is_vista(self):
        """
        Check if current platform is Windows Vista or Windows Server 2008.

        @return: C{True} if the current platform has been detected as Vista
        @rtype: C{bool}
        """
        if getattr(sys, "getwindowsversion", None) is not None:
            return sys.getwindowsversion()[0] == 6
        else:
            return False

    def supportsThreads(self):
        """Can threads be created?
        """
        try:
            return imp.find_module('thread')[0] is None
        except ImportError:
            return False

    @property
    def has_multiProcessSocket(self):
        ''':rtype: a boolean indicating if support for multiprocess
 sockets is available.
        '''
        return hasattr(socket, 'fromfd')



'''Test twisted integration'''
import pulsar
from pulsar import is_failure, is_async
from pulsar.utils.pep import to_bytes, to_string
from pulsar.apps.test import unittest, dont_run_with_thread

from .manage import twisted, mail_client
        
    
@unittest.skipUnless(twisted, 'Requires twisted and a config file')
class TestWebMail(unittest.TestCase):
    concurrency = 'thread'
    server = None
    
    def testMailCient(self):
        client = yield mail_client(timeout=5)
        self.assertTrue(client)


import os
import sys
from time import time

import pulsar
from pulsar.utils.structures import AttributeDictionary
from pulsar.utils.pep import iteritems, itervalues, range

from . import proxy
from .actor import Actor, send
from .defer import async, NOT_DONE
from .concurrency import concurrency
from .mailbox import MonitorMailbox
from .consts import *


__all__ = ['Monitor', 'PoolMixin']


def _spawn_actor(cls, monitor, cfg=None, name=None, aid=None, **kw):
    # Internal function which spawns a new Actor and return its
    # ActorProxyMonitor.
    # *cls* is the Actor class
    # *monitor* can be either the ariber or a monitor
    kind = None
    if issubclass(cls, PoolMixin):
        kind = 'monitor'
    if cfg is None:
        if monitor:
            cfg = monitor.cfg.copy()
        else:
            cfg = pulsar.Config()
    if monitor:
        params = monitor.actorparams()
        name = params.pop('name', name)
        aid = params.pop('aid', aid)
    else:
        if kind != 'monitor':
            raise TypeError('class %s not a valid monitor' % cls)
        params = {}
    for key, value in iteritems(kw):
        if key in cfg.settings:
            cfg.set(key, value)
        else:
            params[key] = value
    #
    if monitor:
        if not kind:
            if not issubclass(cls, Actor):
                raise TypeError('Class %s not a valid actor.' % cls)
            kind = cfg.concurrency
    if not kind:
        raise TypeError('Cannot spawn class %s. not a valid concurrency.' % cls)
    actor_proxy = concurrency(kind, cls, monitor, cfg, name=name,
                              aid=aid, **params)
    # Add to the list of managed actors if this is a remote actor
    if isinstance(actor_proxy, Actor):
        return actor_proxy
    else:
        actor_proxy.monitor = monitor
        monitor.managed_actors[actor_proxy.aid] = actor_proxy
        deferred = proxy.ActorProxyDeferred(actor_proxy)
        actor_proxy.start()
        return deferred


class PoolMixin(Actor):
    '''Not an actor per se, this is a mixin for :class:`Actor`
which manages a pool (group) of actors. Given an :attr:`actor_class`
it makes sure there are always :attr:`cfg.workers` alive.
It is used by both the :class:`Arbiter` and the :class:`Monitor` classes.

.. attribute:: managed_actors

    dictionary with keys given by actor's ids and values by
    :class:`ActorProxyMonitor` instances. These are the actors managed by the
    pool.

'''
    CLOSE_TIMEOUT = 30000000000000
    actor_class = Actor
    '''The class derived form :class:`Actor` which the monitor manages
during its life time.

    Default: :class:`Actor`'''

    def __init__(self, impl):
        super(PoolMixin, self).__init__(impl)
        self.managed_actors = {}
        self.terminated_actors = []
        self.actor_class = self.params.pop('actor_class') or self.actor_class

    def hand_shake(self, *args):
        #Monitors don't do hand shakes
        try:
            self.state = ACTOR_STATES.RUN
            self.fire_event('start')
            self.periodic_task()
        except Exception:
            self.stop(sys.exc_info())
    
    def get_actor(self, aid):
        aid = getattr(aid, 'aid', aid)
        if aid == self.aid:
            return self
        elif aid in self.managed_actors:
            return self.managed_actors[aid]
        elif self.monitor and aid==self.monitor.aid:
            return self.monitor
        
    def active(self):
        return self.running()

    def spawn(self, actor_class=None, **params):
        '''Spawn a new :class:`Actor` and return its
:class:`ActorProxyMonitor`.'''
        actor_class = actor_class or self.actor_class
        return _spawn_actor(actor_class, self, **params)

    def actorparams(self):
        '''Return a dictionary of parameters to be passed to the
spawn method when creating new actors.'''
        return dict(self.params)

    def _remove_actor(self, actor, log=True):
        if log:
            self.logger.info('Removing %s', actor)
        self.managed_actors.pop(actor.aid, None)
        if self.monitor:
            self.monitor._remove_actor(actor, False)
                
    def manage_actors(self, stop=False):
        '''Remove :class:`Actor` which are not alive from the
:class:`PoolMixin.managed_actors` and return the number of actors still alive.

:parameter stop: if ``True`` stops all alive actor.
'''
        alive = 0
        if self.managed_actors:
            for aid, actor in list(iteritems(self.managed_actors)):
                alive += self.manage_actor(actor, stop)
        return alive

    def manage_actor(self, actor, stop=False):
        '''If an actor failed to notify itself to the arbiter for more than
the timeout. Stop the arbiter.'''
        if not self.running():
            stop = True
        if not actor.is_alive():
            if not actor.should_be_alive() and not stop:
                return 1
            actor.join(ACTOR_TERMINATE_TIMEOUT)
            self._remove_actor(actor)
            return 0
        timeout = None
        started_stopping = bool(actor.stopping_start)
        stop = stop or started_stopping
        if not stop and actor.notified:
            gap = time() - actor.notified
            stop = timeout = gap > actor.cfg.timeout
        if stop:   # we are shutting down
            if not actor.mailbox or actor.should_terminate():
                if not actor.mailbox:
                    self.logger.info('Terminating %s. No mailbox.', actor)
                else:
                    self.logger.warning('Terminating %s. Timeout.', actor)
                actor.terminate()
                self.terminated_actors.append(actor)
                self._remove_actor(actor)
                return 0
            elif not started_stopping:
                if timeout:
                    self.logger.warning('Stopping %s. Timeout', actor)
                else:
                    self.logger.info('Stopping %s.', actor)
                self.send(actor, 'stop')
        return 1

    def spawn_actors(self):
        '''Spawn new actors if needed. If the :class:`PoolMixin` is spawning
do nothing.'''
        to_spawn = self.cfg.workers - len(self.managed_actors)
        if self.cfg.workers and to_spawn > 0:
            for _ in range(to_spawn):
                self.spawn()

    def stop_actors(self):
        """Maintain the number of workers by spawning or killing
as required."""
        if self.cfg.workers:
            num_to_kill = len(self.managed_actors) - self.cfg.workers
            for i in range(num_to_kill, 0, -1):
                w, kage = 0, sys.maxsize
                for worker in itervalues(self.managed_actors):
                    age = worker.impl.age
                    if age < kage:
                        w, kage = w, age
                self.manage_actor(w, True)
    
    @async()
    def close_actors(self):
        '''Close all managed :class:`Actor`.'''
        # Stop all of them
        to_stop = self.manage_actors(stop=True)
        while to_stop:
            yield NOT_DONE
            to_stop = self.manage_actors(stop=True)
    
    
class Monitor(PoolMixin):
    '''A monitor is a **very** special :class:`Actor` and :class:`PoolMixin`
which shares the same :class:`EventLoop` with the :class:`Arbiter` and
therefore lives in the main thread of the  process domain.
The Arbiter manages monitors which in turn manage a set of :class:`Actor`
performing similar tasks.

In other words, you may have a monitor managing actors for serving HTTP
requests on a given port, another monitor managing actors consuming tasks
from a task queue and so forth. You can think of :class:`Monitor` as
managers of pools of :class:`Actor`.

Monitors are created by invoking the :meth:`Arbiter.add_monitor`
functions and not by directly invoking the constructor. Therefore
adding a new monitor to the arbiter follows the pattern::

    import pulsar

    m = pulsar.arbiter().add_monitor(pulsar.Monitor, 'mymonitor')
'''
    socket = None

    @property
    def cpubound(self):
        return False
    
    @property
    def arbiter(self):
        return self.monitor

    def is_process(self):
        return False

    def is_monitor(self):
        return True

    def monitor_task(self):
        '''Monitor specific task called by the :meth:`Monitor.periodic_task`.
By default it does nothing. Override if you need to.'''
        pass

    def periodic_task(self):
        '''Overrides the :meth:`Actor.on_task`
:ref:`actor callback <actor-callbacks>` to perform
the monitor :class:`EventLoop` tasks, which are:

* To maintain a responsive set of actors ready to perform their duty.
* To perform its own tasks.

The implementation goes as following:

* It calls :meth:`PoolMixin.manage_actors` which removes from the live
  actors dictionary all actors which are not alive.
* Spawn new actors if required by calling :meth:`PoolMixin.spawn_actors`
  and :meth:`PoolMixin.stop_actors`.
* Call :meth:`Monitor.monitor_task` which performs the monitor specific
  task.

Users shouldn't need to override this method, but use
:meth:`Monitor.monitor_task` instead.'''
        if self.running():
            self.manage_actors()
            self.spawn_actors()
            self.stop_actors()
            self.monitor_task()
        self.event_loop.call_soon(self.periodic_task)

    # HOOKS
    @async(max_errors=None)
    def _stop(self):
        yield self.close_actors()
        if not self.terminated_actors:
            self.monitor._remove_actor(self)
        self.fire_event('stop')

    @property
    def multithread(self):
        return self.cfg.concurrency == 'thread'

    @property
    def multiprocess(self):
        return self.cfg.concurrency == 'process'
    
    @property
    def requestloop(self):
        return self.monitor.requestloop

    def info(self):
        data = {'actor': {'actor_class':self.actor_class.__name__,
                          'concurrency':self.cfg.concurrency,
                          'name':self.name,
                          'age':self.impl.age,
                          'workers': len(self.managed_actors)}}
        if not self.started():
            return data
        data['workers'] = [a.info for a in itervalues(self.managed_actors)\
                           if a.info]
        return data

    def proxy_mailbox(address):
        return self.arbiter.proxy_mailboxes.get(address)

    def get_actor(self, aid):
        #Delegate get_actor to the arbiter
        a = super(Monitor, self).get_actor(aid)
        if a is None:
            a = self.monitor.get_actor(aid)
        return a

    # OVERRIDES INTERNALS        
    def _setup_event_loop(self):
        self.mailbox = self._mailbox()
        
    def _run(self):
        pass
    
    def _mailbox(self):
        return MonitorMailbox(self)



from time import time

from pulsar import CommandError

from .defer import is_async
from .proxy import command, ActorProxyMonitor
from .consts import NOT_DONE


@command()
def ping(request):
    return 'pong'

@command()
def echo(request, message):
    '''Returns *message*'''
    return message

@command()
def config(request, setget, name, *value):
    setget = setget.lower()
    if setget == 'get':
        if len(values) > 0:
            raise CommandError('"config get" accept only one parameter')
        return client.actor.cfg.get(name)
    elif setget == 'set':
        if len(values) > 1:
            raise CommandError('"config get" accept only two parameters')
        client.actor.cfg.set(name, value[0])
    else:
        raise CommandError('config must be followed by set or get')

@command()
def run(request, callable, *args, **kwargs):
    '''Execute a python *callable*.'''
    return callable(request.actor, *args, **kwargs)

@command(ack=False)
def stop(request):
    '''Stop the actor from running.'''
    return request.actor.stop()
    
@command()
def notify(request, info):
    '''The actor notify itself with a dictionary of information.
The command perform the following actions:

* Update the mailbox to the current consumer of the actor connection
* Update the info dictionary
* Returns the time of the update
'''
    t = time()
    remote_actor = request.caller
    if isinstance(remote_actor, ActorProxyMonitor):
        remote_actor.mailbox = request.connection.current_consumer
        info['last_notified'] = t
        remote_actor.info = info
        callback = remote_actor.callback
        # if a callback is still available, this is the first
        # time we got notified
        if callback:
            remote_actor.callback = None
            callback.callback(remote_actor)
    return t
    
@command()
def spawn(request, **kwargs):
    '''Spawn a new actor.'''
    return request.actor.spawn(**kwargs)

@command()
def info(request):
    ''' Returns information and statistics about the server as a json string'''
    return request.actor.info()
     
@command()
def kill_actor(request, aid):
    '''Kill an actor with id ``aid``. This command can only be executed by the
arbiter, therefore a valid sintax is only::

    send('arbiter', 'kill_actor', 'abc')

Return 'stopped abc` if succesful, otherwise it returns ``None``.
'''
    arb = request.actor 
    if arb.is_arbiter():
        arb.send(aid, 'stop')
        while arb.get_actor(aid):
            yield NOT_DONE
        yield 'stopped %s' % aid



"""\
Security related helpers such as secure password hashing tools.
"""
from hashlib import sha1
from uuid import uuid4
import string
from random import SystemRandom

from .httpurl import range, ascii_letters

SALT_CHARS = ascii_letters + string.digits


_sys_rng = SystemRandom()

def gen_unique_id():
    return str(uuid4())

def gen_salt(length):
    """Generate a random string of SALT_CHARS with specified ``length``."""
    if length <= 0:
        raise ValueError('requested salt of length <= 0')
    return ''.join(_sys_rng.choice(SALT_CHARS) for _ in range(length))

def _hash_internal(salt, password):
    return sha1(('%s%s' % (salt, password)).encode('utf-8')).hexdigest()

def generate_password_hash(password, salt_length=8):
    salt = gen_salt(salt_length)
    h = _hash_internal(salt, password)
    return '%s$%s' % (salt, h)

def check_password_hash(pwhash, password):
    if pwhash.count('$') != 1:
        return False
    salt, hashval = pwhash.split('$')
    return _hash_internal(salt, password) == hashval



'''
A test plugin for profiling test cases and generating Html reports. It uses
the ``cProfiler`` from the standard libraries.

.. autoclass:: Profile
   :members:
   :member-order: bysource
   
'''
import os
import re
import time
import shutil
import tempfile
import cProfile as profiler
import pstats
from datetime import datetime

import pulsar
from pulsar.utils.httpurl import ispy3k
from pulsar.apps import test

if ispy3k:
    from io import StringIO as Stream
else:   #pragma    nocover
    from io import BytesIO as Stream

other_filename = 'unknown'
line_func = re.compile(r'(?P<line>\d+)\((?P<func>\w+)\)')
template_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'htmlfiles','profile')
headers = (
    ('ncalls',
     'total number of calls'),
    ('primitive calls',
     'Number primitive calls (calls not induced via recursion)'),
    ('tottime',
     'Total time spent in the given function (excluding time spent '\
     'in calls to sub-functions'),
    ('percall',
     'tottime over ncalls, the time spent by each call'),
    ('cumtime',
     'Total time spent in the given function including all subfunctions'),
    ('percall',
     'cumtime over primitive calls'),
    ('function',''),
    ('lineno',''),
    ('filename','')
    )

def absolute_file(val):
    dir = os.getcwd()
    return os.path.join(dir,val)

def make_stat_table(data):
    yield "<thead>\n<tr>\n"
    for head, description in headers:
        yield '<th title="{1}">{0}</th>'.format(head, description)
    yield '\n</tr>\n</thead>\n<tbody>\n'
    for row in data:
        yield '<tr>\n'
        for col in row:
            yield '<td>{0}</td>'.format(col)
        yield '\n</tr>\n'
    yield '</tbody>'

def data_stream(lines, num=None):
    if num:
        lines = lines[:num]
    for line in lines:
        if not line:
            continue
        fields = [field for field in line.split() if field is not '']
        if len(fields) == 6:
            valid = True
            new_fields = fields[0].split('/')
            if len(new_fields) == 1:
                new_fields.append(new_fields[0])
            for f in fields[1:-1]:
                try:
                    float(f)
                except Exception:
                    valid = False
                    break
                new_fields.append(f)
            if not valid:
                continue
            filenames = fields[-1].split(':')
            linefunc = filenames.pop()
            match = line_func.match(linefunc)
            if match:
                lineno, func = match.groups()
                filename = ':'.join(filenames)
                filename = filename.replace('\\','/')
                new_fields.extend((func, lineno, filename))
            else:
                new_fields.extend(('', '', other_filename))
            yield new_fields

def copy_file(filename, target, context=None):
    with open(os.path.join(template_path,filename),'r') as file:
        stream = file.read()
    if context:
        stream = stream.format(context)
    with open(os.path.join(target,filename),'w') as file:
        file.write(stream)


class Profile(test.TestPlugin):
    ''':class:`pulsar.apps.test.TestPlugin` for profiling test cases.'''
    desc = '''Profile tests using the cProfile'''
    profile_stats_path = pulsar.Setting(
                                flags=['--profile-stats-path'],
                                default='htmlprof',
                                desc='''location of profile directory''',
                                validator=absolute_file)

    def configure(self, cfg):
        self.config = cfg
        self.profile_stats_path = cfg.profile_stats_path
        dir, name = os.path.split(self.profile_stats_path)
        fname = '.'+name
        self.profile_temp_path = os.path.join(dir, fname)

    def before_test_function_run(self, test, local):
        # If active return a TestProfile instance wrapping the real test case.
        if self.config.profile:
            local.prof = profiler.Profile()
            local.tmp = tempfile.mktemp(dir=self.profile_temp_path)
            local.prof.enable()
        
    def after_test_function_run(self, test, local, result, async):
        if self.config.profile:
            local.prof.disable()
            local.prof.dump_stats(local.tmp)

    def on_start(self):
        if self.config.profile:
            self.remove_dir(self.profile_temp_path, build=True)

    def remove_dir(self, dir, build=False):
        sleep = 0
        if os.path.exists(dir):
            shutil.rmtree(dir)
            sleep = 0.2
        if build:
            time.sleep(sleep)
            os.mkdir(dir)

    def on_end(self):
        if self.config.profile:
            files = [os.path.join(self.profile_temp_path,file) for file in\
                     os.listdir(self.profile_temp_path)]
            if not files:
                return
            stats = pstats.Stats(*files, **{'stream': Stream()})
            stats.sort_stats('time', 'calls')
            stats.print_stats()
            stats_str = stats.stream.getvalue()
            self.remove_dir(self.profile_temp_path)
            stats_str = stats_str.split('\n')
            run_info = 'Executed %s.' % datetime.now().isoformat()
            for n, line in enumerate(stats_str):
                b = 0
                while b < len(line) and line[b] == ' ':
                    b += 1
                line = line[b:]
                if line:
                    if line.startswith('ncalls'):
                        break
                    bits = line.split(' ')
                    try:
                        ncalls = int(bits[0])
                    except Exception:
                        continue
                    else:
                        run_info += ' ' + line
            data = ''.join(make_stat_table(data_stream(stats_str[n+1:], 100)))
            self.remove_dir(self.profile_stats_path, build=True)
            for file in os.listdir(template_path):
                if file == 'index.html':
                    copy_file(file, self.profile_stats_path,
                              {'table': data,
                               'run_info': run_info,
                               'version': pulsar.__version__})
                else:
                    copy_file(file, self.profile_stats_path)



#An example on how to use Server Hooks in a config file 

requests = {}

def _pre_request(worker, request):
    pass

def _post_request(worker, response):
    environ = response.environ
    connection = response.connection
    user_agent = environ['HTTP_USER_AGENT']
    path = environ['PATH_INFO']
    address = environ['REMOTE_ADDR']
    session = connection.session
    request_in_session = connection.processed
    pass
    


import time
from multiprocessing import Process, current_process
from threading import Thread, current_thread

from pulsar import system, platform
from pulsar.utils.security import gen_unique_id

from .proxy import ActorProxyMonitor, get_proxy
from .access import get_actor, get_actor_from_id
from .threads import KillableThread


__all__ = ['Concurrency', 'concurrency']

def concurrency(kind, actor_class, monitor, cfg, **params):
    '''Function invoked by the :class:`Arbiter` or a :class:`Monitor` when
spawning a new :class:`Actor`. It created a :class:`Concurrency` instance
which handle the contruction and the lif of an :class:`Actor`.

:paramater kind: Type of concurrency
:paramater monitor: The monitor (or arbiter) managing the :class:`Actor`.
:rtype: a :class:`Councurrency` instance
'''
    if kind == 'monitor':
        c = MonitorConcurrency()
    elif kind == 'thread':
        c = ActorThread()
    elif kind == 'process':
        c = ActorProcess()
    else:
        raise ValueError('Concurrency %s not supported in pulsar' % kind)
    return c.make(kind, actor_class, monitor, cfg, **params)


class Concurrency(object):
    '''Actor :class:`Concurrency` is responsible for the actual spawning of
actors according to a concurrency implementation. Instances are pickable
and are shared between the :class:`Actor` and its
:class:`ActorProxyMonitor`.

:parameter concurrency: string indicating the concurrency implementation.
    Valid choices are ``monitor``, ``process``, ``thread``.
:parameter actor_class: :class:`Actor` or one of its subclasses.
:parameter timeout: timeout in seconds for the actor.
:parameter kwargs: additional key-valued arguments to be passed to the actor
    constructor.
'''
    _creation_counter = 0
    def make(self, kind, actor_class, monitor, cfg, name=None, aid=None, **kw):
        self.__class__._creation_counter += 1
        self.aid = aid or gen_unique_id()[:8]
        self.age = self.__class__._creation_counter
        self.name = name or actor_class.__name__.lower()
        self.kind = kind
        self.cfg = cfg
        self.actor_class = actor_class
        self.params = kw
        self.params['monitor'] = monitor
        return self.get_actor()

    @property
    def unique_name(self):
        return '%s(%s)' % (self.name, self.aid)

    def __repr__(self):
        return self.unique_name
    __str__ = __repr__

    def get_actor(self):
        self.daemon = True
        self.params['monitor'] = get_proxy(self.params['monitor'])
        return ActorProxyMonitor(self)


class MonitorConcurrency(Concurrency):
    ''':class:`Concurrency` class for monitors such as the :class:`Arbiter`
and :class:`Monitor`. Monitors live in the **mainthread** of the master process
and therefore do not require to be spawned.'''
    def get_actor(self):
        return self.actor_class(self)

    def start(self):
        pass

    def is_active(self):
        return self.actor.is_alive()

    @property
    def pid(self):
        return current_process().pid


class ActorConcurrency(Concurrency):
    '''Base class for all :class:`Actor` concurrency models. Must implement
the **start** method.'''
    def run(self):
        actor = self.actor_class(self)
        actor.start()


class ActorProcess(ActorConcurrency, Process):
    '''Actor on a process'''


class ActorThread(ActorConcurrency, KillableThread):
    '''Actor on a thread'''

    def terminate(self, kill=False):
        if self.is_alive():
            actor = get_actor_from_id(self.aid)
            me = get_actor()
            if not kill and actor and actor != me:
                actor.stop(1)
                me.event_loop.call_later(1, self.terminate, True)
            else:
                KillableThread.terminate(self)



# -*- coding: utf-8 -*-
'''WebSocket_ Protocol :class:`Frame` and :class:`FrameParser` classes.
These two classes can be used for both clients and server protocols.

.. _WebSocket: http://tools.ietf.org/html/rfc6455'''
import array
import os
from struct import pack, unpack
from array import array
from io import BytesIO

from .pep import ispy3k, range, to_bytes, is_string
from .exceptions import ProtocolError

DEFAULT_VERSION = 13
SUPPORTED_VERSIONS = (DEFAULT_VERSION,)
WS_EXTENSIONS = {}
WS_PROTOCOLS = {}

if ispy3k:
    i2b = lambda n : bytes((n,))
    
    def is_text_data(data):
        return isinstance(data, str)
else: # pragma : nocover
    i2b = lambda n : chr(n)
    
    def is_text_data(data):
        return True

def int2bytes(*ints):
    '''convert a series of integers into bytes'''
    return b''.join((i2b(i) for i in ints))

def get_version(version):
    try:
        version = int(version or DEFAULT_VERSION)
    except Exception:
        pass
    if version not in SUPPORTED_VERSIONS:
        raise ProtocolError('Version %s not supported.' % version)
    return version

    
class Extension(object):
    
    def receive(self, data):
        return data
        
    def send(self, data):
        return data
    
    
class Frame(object):
    """Implements the framing protocol as defined by
specification supporting protocol version 13::
    
    >>> f = Frame('hello world', final=1)
    >>> data = f.build()
    >>> f = Frame()
    >>> f.parser.send(bytes[1])
    >>> f.parser.send(bytes[2])
    >>> f.parser.send(bytes[2:])

.. attribute:: version

    The version of the frame protocol.
        
.. attribute:: masking_key

    The frame masking key used for encoding and decoding data.
"""
    __slots__ = ('fin', 'opcode', 'masking_key', 'rsv1', 'rsv2', 'rsv3',
                 'msg', 'version', 'payload_length', 'body')
    def __init__(self, message=None, opcode=None, version=None,
                 masking_key=None, final=False, rsv1=0, rsv2=0, rsv3=0,
                 build_frame=True):
        self.version = get_version(version)
        if opcode is None and message is not None:
            opcode = 0x1 if is_text_data(message) else 0x2
        message = to_bytes(message or b'')
        self.payload_length = len(message)
        if opcode is None:
            raise ProtocolError('WEBSOCKET opcode not available')
        self.version = version
        self.opcode = opcode
        if masking_key:
            masking_key = to_bytes(masking_key)
            if len(masking_key) != 4:
                raise ProtocolError('WEBSOCKET masking key must be 4 bytes long')
        self.masking_key = masking_key
        self.fin = 0x1 if final else 0
        self.rsv1 = rsv1
        self.rsv2 = rsv2
        self.rsv3 = rsv3
        if build_frame:
            self.body = message
            self.msg = self._build_frame(message)
        else:
            self.body = None
            self.msg = bytearray(message)
        
    @property
    def final(self):
        return bool(self.fin)
    
    @property
    def masked(self):
        return bool(self.masking_key)
    
    @property
    def is_data(self):
        return self.opcode in (0x1, 0x2)
    
    @property
    def is_close(self):
        return self.opcode == 0x8
    
    def _build_frame(self, message):
        #Builds a frame from the instance's attributes
        header = BytesIO()
        if 0x3 <= self.opcode <= 0x7 or 0xB <= self.opcode:
            raise ProtocolError('WEBSOCKET opcode cannot be a reserved opcode')
        ## +-+-+-+-+-------+
        ## |F|R|R|R| opcode|
        ## |I|S|S|S|  (4)  |
        ## |N|V|V|V|       |
        ## | |1|2|3|       |
        ## +-+-+-+-+-------+
        header.write(i2b(((self.fin << 7)
                       | (self.rsv1 << 6)
                       | (self.rsv2 << 5)
                       | (self.rsv3 << 4)
                       | self.opcode)))
        ##                 +-+-------------+-------------------------------+
        ##                 |M| Payload len |    Extended payload length    |
        ##                 |A|     (7)     |             (16/63)           |
        ##                 |S|             |   (if payload len==126/127)   |
        ##                 |K|             |                               |
        ## +-+-+-+-+-------+-+-------------+ - - - - - - - - - - - - - - - +
        ## |     Extended payload length continued, if payload len == 127  |
        ## + - - - - - - - - - - - - - - - +-------------------------------+
        if self.masking_key:
            mask_bit = 1 << 7
        else:
            mask_bit = 0
        length = self.payload_length 
        if length < 126:
            header.write(i2b(mask_bit | length))
        elif length < (1 << 16):
            header.write(i2b(mask_bit | 126))
            header.write(pack('!H', length))
        elif length < (1 << 63):
            header.write(i2b(mask_bit | 127))
            header.write(pack('!Q', length))
        else:
            raise ProtocolError('WEBSOCKET frame too large')
        ## + - - - - - - - - - - - - - - - +-------------------------------+
        ## |                               |Masking-key, if MASK set to 1  |
        ## +-------------------------------+-------------------------------+
        ## | Masking-key (continued)       |          Payload Data         |
        ## +-------------------------------- - - - - - - - - - - - - - - - +
        ## :                     Payload Data continued ...                :
        ## + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        ## |                     Payload Data continued ...                |
        ## +---------------------------------------------------------------+
        if not self.masking_key:
            header.write(message)
        else:
            header.write(self.masking_key)
            header.write(self.mask(message))
        return header.getvalue()
       
    def mask(self, data):
        '''Performs the masking or un-masking operation on data using the
simple masking algorithm::
    
    j = i MOD 4
    transformed-octet-i = original-octet-i XOR masking-key-octet-j
    
This method is invoked when encoding/decoding frames with a :attr:`masking_key`
attribute set.'''
        mask_size = len(self.masking_key)
        key = array('B', self.masking_key)
        data = array('B', data)
        for i in range(len(data)):
            data[i] ^= key[i % mask_size]
        return data.tobytes() if ispy3k else data.tostring()
    unmask = mask

    
class FrameParser(object):
    '''Decoder and encoder for the websocket protocol.
    
.. attribute:: version

    Optional protocol version (Default 13).
    
.. attribute:: kind

    * 0 for parsing client's frames and sending server frames (to be used
      in the server)
    * 1 for parsing server frames and sending client frames (to be used
      by the client)
    * 2 Assumes always unmasked data
'''    
    def __init__(self, version=None, kind=0, extensions=None, protocols=None):
        self.version = get_version(version)
        self._ext_middleware, self._extensions =\
            self.ws_middleware(extensions, WS_EXTENSIONS)
        self._pro_middleware, self._protocols =\
            self.ws_middleware(extensions, WS_PROTOCOLS)
        self._frame = None  #current frame
        self._buf = None
        self._kind = kind

    def ws_middleware(self, names, group):
        mw = []
        av = []
        if names:
            for name in names:
                if name in group:
                    av.append(name)
                    mw.append(group[name]())
        return mw, tuple(av)
    
    @property
    def kind(self):
        return self._kind
    
    @property
    def extensions(self):
        return self._extensions
    
    @property
    def protocols(self):
        return self._protocols
    
    @property
    def masked(self):
        return self.kind == 1
    
    @property
    def expect_masked(self):
        return True if self.kind == 0 else False
    
    def decode(self, data):
        return self.execute(data), bytearray()
    
    def encode(self, data, final=True, masking_key=None, **params):
        '''Encode data into a :class:`Frame`.
        
:parameter final: Indicating if this is a final Frame.
:parameter masking_key: Optional making key used only if :attr:`kind` is 1 
    (Client frames).
    '''
        if self.masked:
            masking_key = masking_key or os.urandom(4)
            return Frame(data, masking_key=masking_key, final=final, **params)
        else:
            return Frame(data, final=final, **params)
    
    def ping(self, body=None):
        '''return a `ping` :class:`Frame`.'''
        return self.encode(body, opcode=0x9)
    
    def pong(self, body=None):
        '''return a `pong` :class:`Frame`.'''
        return self.encode(body, opcode=0xA)
    
    def close(self, body=None):
        '''return a `close` :class:`Frame`.'''
        return self.encode(body, opcode=0x8)
    
    def continuation(self, body=None, final=True):
        '''return a `continuation` :class:`Frame`.'''
        return self.encode(body, opcode=0, final=final)
    
    def replay_to(self, frame):
        '''Build a :class:`Frame` as a reply to *frame*.'''
        if frame.opcode == 0x9:
            return self.pong(frame.body)
        
    def decode(self, data=None):
        '''Decode bytes data into a :class:`Frame`. If :attr:`kind` is 0
it decodes into a client frame (masked frame) while if is 1 or 2 it decodes
into a server frame (unmasked).'''
        if data:
            if self._buf:
                data = self._buf + data
        else:
            data = self._buf
        if not data: 
            return None
        masked_frame = self.expect_masked
        frame = self._frame
        # No opcode yet
        if frame is None:
            if len(data) < 2:
                return self.save_buf(frame, data)
            first_byte, second_byte = unpack("BB", data[:2])
            fin = (first_byte >> 7) & 1
            rsv1 = (first_byte >> 6) & 1
            rsv2 = (first_byte >> 5) & 1
            rsv3 = (first_byte >> 4) & 1
            opcode = first_byte & 0xf
            if fin not in (0, 1):
                raise WebSocketProtocolError('FIN must be 0 or 1')
            is_masked = bool(second_byte & 0x80)
            if masked_frame != is_masked:
                if masked_frame:
                    raise ProtocolError('WEBSOCKET unmasked client frame.')
                else:
                    raise ProtocolError('WEBSOCKET masked server frame.')
            payload_length = second_byte & 0x7f
            # All control frames MUST have a payload length of 125 bytes or less
            if opcode > 0x7 and payload_length > 125:
                raise ProtocolError('WEBSOCKET frame too large')
            msg, data = data[:2], data[2:]
            frame = Frame(msg, opcode, self.version, final=fin,
                          rsv1=rsv1, rsv2=rsv2, rsv3=rsv3, build_frame=False)
            frame.payload_length = payload_length
        #   
        if frame.masking_key is None:
            # All control frames MUST have a payload length of 125 bytes or less
            d = None
            mask_length = 4 if masked_frame else 0
            if frame.payload_length == 0x7e: #126
                if len(data) < 2 + mask_length: # 2 + 4 for mask
                     return self.save_buf(frame, data)
                d, data = data[:2] , data[2:]
                frame.payload_length = unpack("!H", d)[0]
            elif frame.payload_length == 0x7f: #127
                if len(data) < 8 + mask_length:  # 8 + 4 for mask
                     return self.save_buf(frame, data)
                d, data = data[:8] , data[8:]
                frame.payload_length = unpack("!Q", d)[0]
            elif len(data) < mask_length:
                return self.save_buf(frame, data)
            if d:
                frame.msg.extend(d)
            if mask_length:
                frame.masking_key, data = data[:mask_length], data[mask_length:]
            else:
                frame.masking_key = b''
            frame.msg.extend(frame.masking_key)
                
        if len(data) < frame.payload_length:
            self.save_buf(frame, data)
        # We have a frame
        else:
            payload = data[:frame.payload_length] # payload data
            frame.msg.extend(payload)
            self.save_buf(None, data[frame.payload_length:])
            for extension in self.extensions:
                data = extension.receive(frame, data)
            if frame.masking_key:
                data = frame.unmask(data)
            if frame.opcode == 0x1:
                data = data.decode("utf-8", "replace")
            frame.body = data
            return frame
            
    def save_buf(self, frame, data):
        self._frame = frame
        self._buf = data
        


'''The :mod:`pulsar.apps.pubsub` implements a middleware
handler for the Publish/Subscribe pattern.

.. epigraph::

    In software architecture, publish/subscribe is a messaging pattern where
    senders of messages, called publishers, do not program the messages to
    be sent directly to specific receivers, called subscribers.
    Instead, published messages are characterized into classes, without
    knowledge of what, if any, subscribers there may be. Similarly, subscribers
    express interest in one or more classes, and only receive messages that are
    of interest, without knowledge of what, if any, publishers there are.
    
    -- wikipedia_


When using this middleware, one starts by creating a :class:`PubSub` handler::

    from pulsar.apps.pubsub import PubSub
    
    pubsub = PubSub(connection_string=None, ...)
    
The ``connection_string`` parameter is needed in order to select the backend
to use. If not supplied, the default ``local://`` backend is used.


PubSub handler
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: PubSub
   :members:
   :member-order: bysource
   
   
PubSub backend
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: PubSubBackend
   :members:
   :member-order: bysource 
   
   
.. _wikipedia: http://en.wikipedia.org/wiki/Publish%E2%80%93subscribe_pattern'''
import logging

import pulsar
from pulsar import get_actor
from pulsar.utils.pep import to_string
from pulsar.utils.log import local_property


LOGGER = logging.getLogger('pulsar.pubsub') 


class Client(object):
    '''Interface for a client of :class:`PubSub` handler. Instances of this
:class:`Client` are callable object and are called once a new message has
arrived from a subscribed channel. The callable accepts two parameters:

* ``channels`` the channels which originated the message
* ``message`` the message
'''
    def __call__(self, channel, message):
        raise NotImplementedError
        

class PubSub(object):
    '''Publish/Subscribe paradigm handler.

.. attribute:: backend

    The :class:`PubSubBackend` for this handler.
        
.. attribute:: encoder

    Optional callable which encode the messages before they are published.
    
.. attribute:: clients

    Set of all clients for this :class:`PubSub` handler.
'''
    def __init__(self, backend=None, encoder=None, **params):
        be = PubSubBackend.make(backend=backend, **params)
        self.backend = PubSubBackend.get(be.id, backend=be)
        self.encoder = encoder
    
    @property
    def clients(self):
        return self.backend.clients
    
    @property
    def id(self):
        return self.backend.id
    
    @property
    def name(self):
        return self.backend.name
    
    def add_client(self, client):
        '''Add a new ``client`` to the set of all :attr:`clients`. Clients
must have the ``write`` method available. When a new message is received
from the publisher, the :meth:`broadcast` method will notify all
:attr:`clients` via the ``write`` method.'''
        self.backend.add_client(client)
        
    def remove_client(self, client):
        '''Remove *client* from the set of all :attr:`clients`.'''
        self.backend.remove_client(client)
        
    def publish(self, channel, message):
        '''Publish a ``message`` to ``channel``. It invokes the
:meth:`PubSubBackend.publish` method after the message has been encoded
via the :attr:`encoder` callable (if available).'''
        if self.encoder:
            message = self.encoder(message)
        return self.backend.publish(channel, message)
    
    def subscribe(self, *channels):
        '''Subscribe to the server which publish messages. Must be
implemented by subclasses.'''
        return self.backend.subscribe(*channels)
    
    def unsubscribe(self, *channels):
        '''Un-subscribe to the server which publish messages. Must be
implemented by subclasses.'''
        return self.backend.unsubscribe(*channels)
    
    def close(self):
        '''Close connections'''
        return self.backend.close()


class PubSubBackend(pulsar.Backend):
    '''Publish/Subscribe Backend interface.
    
.. attribute:: clients

    Set of all clients for this :class:`PubSub` handler.
'''
    @local_property
    def clients(self):
        '''The set of clients for this :class:`PubSub` handler.'''
        return set()
    
    @local_property
    def lock(self):
        return Lock()
    
    @classmethod
    def path_from_scheme(cls, scheme):
        return 'pulsar.apps.pubsub.%s' % scheme
    
    def add_client(self, client):
        '''Add a new ``client`` to the set of all :attr:`clients`. Clients
must have the ``write`` method available. When a new message is received
from the publisher, the :meth:`broadcast` method will notify all
:attr:`clients` via the ``write`` method.'''
        self.clients.add(client)
        
    def remove_client(self, client):
        '''Remove *client* from the set of all :attr:`clients`.'''
        self.clients.discard(client)
        
    def publish(self, channel, message):
        '''Publish a ``message`` into ``channel``. Must be implemented
by subclasses.'''
        raise NotImplementedError
    
    def subscribe(self, *channels):
        '''Subscribe to the server which publish messages. Must be
implemented by subclasses.'''
        raise NotImplementedError
    
    def unsubscribe(self, *channels):
        '''Un-subscribe to the server which publish messages. Must be
implemented by subclasses.'''
        raise NotImplementedError
        
    def broadcast(self, channel, message):
        '''Broadcast ``message`` to all :attr:`clients`.'''
        remove = set()
        channel = to_string(channel)
        message = self.decode(message)
        clients = tuple(self.clients)
        for client in clients:
            try:
                client(channel, message)
            except Exception as e:
                LOGGER.exception('Exception while processing pub/sub client. '
                                 'Removing it.')
                remove.add(client)
        self.clients.difference_update(remove)
    
    def close(self):
        '''Close this :class:`PubSubBackend`.'''
        actor = get_actor()
        actor.params.pubsub.pop(self.id, None)
        return self.unsubscribe()
    
    def decode(self, message):
        '''Convert message to a string.'''
        return to_string(message)
    
    @classmethod
    def get(cls, id, actor=None, backend=None):
        actor = actor or get_actor()
        if not actor:
            raise RuntimeError('Cannot initialise pubsub when no actor.')
        if 'pubsub' not in actor.params:
            actor.params.pubsub = {}
        be = actor.params.pubsub.get(id)
        if not be and backend:
            be = backend
            actor.params.pubsub[id] = be
        return be
    


# -*- coding: utf-8 -
import sys

import pulsar
from pulsar.apps.wsgi import WSGIServer, LazyWsgi

from django.core.management.base import BaseCommand, CommandError
from django.core.wsgi import get_wsgi_application

PULSAR_OPTIONS = pulsar.make_optparse_options(apps=['socket'])


class Wsgi(LazyWsgi):
    
    def setup(self):
        from django.conf import settings
        return get_wsgi_application()
        
    
class Command(BaseCommand):
    option_list = BaseCommand.option_list + PULSAR_OPTIONS
    help = "Starts a fully-functional Web server using pulsar."
    args = 'pulse --help for usage'

    # Validation is called explicitly each time the server is reloaded.
    requires_model_validation = False

    def handle(self, *args, **options):
        if args:
            raise CommandError('pulse --help for usage')
        callable = Wsgi()
        if options.pop('dryrun', False) == True:
            return callable
        callable.setup()
        WSGIServer(callable=callable, cfg=options, parse_console=False,
                   name=self.app_name()).start()
                   
    def app_name(self):
        '''Used by the test suite to run several applications.'''
        actor = pulsar.get_actor()
        name = None
        if actor:
            name = actor.params.django_pulsar_name
        return name or 'pulsar_django'
        


'''
Documentation for utilities used by pulsar internals. This module is
independent from all other pulsar modules and therefore can be used
as a stand-alone library.

HTTP
============

.. automodule:: pulsar.utils.httpurl

   

.. _tools-ws-parser:

Websocket
==============================

.. automodule:: pulsar.utils.websocket

Frame
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: Frame
   :members:
   :member-order: bysource
   
   
Frame Parser
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: FrameParser
   :members:
   :member-order: bysource
   
   
.. _api-config:

Configuration
==================

.. automodule:: pulsar.utils.config


     
.. _socket_address:

Socket and addresses
=======================

.. automodule:: pulsar.utils.sockets


Internals
=======================


.. module:: pulsar.utils.system

System info
~~~~~~~~~~~~~~~~~

.. autofunction:: system_info



.. module:: pulsar.utils.tools

checkarity
~~~~~~~~~~~~~~~~~

.. autofunction:: checkarity


Structures
==================

.. automodule:: pulsar.utils.structures


HTML & Text
==================

.. automodule:: pulsar.utils.html
'''


from functools import partial

import pulsar
from pulsar import multi_async
from pulsar.utils.pep import range
from pulsar.apps.test import unittest, dont_run_with_thread

from .manage import server, Echo, EchoServerProtocol



class TestEchoServerThread(unittest.TestCase):
    concurrency = 'thread'
    server = None
    
    @classmethod
    def setUpClass(cls):
        s = server(name=cls.__name__.lower(), bind='127.0.0.1:0',
                   backlog=1024, concurrency=cls.concurrency)
        cls.server = yield pulsar.send('arbiter', 'run', s)
        cls.client = Echo(cls.server.address)
        
    @classmethod
    def tearDownClass(cls):
        if cls.server:
            yield pulsar.send('arbiter', 'kill_actor', cls.server.name)
            
    def test_server(self):
        self.assertTrue(self.server)
        self.assertEqual(self.server.callable, EchoServerProtocol)
        self.assertTrue(self.server.address)
        
    def test_ping(self):
        c = self.client
        result = yield c.request(b'ciao luca')
        self.assertEqual(result, b'ciao luca')
        
    def test_large(self):
        '''Echo a 3MB message'''
        msg = b''.join((b'a' for x in range(2**13)))
        result = yield self.client.request(msg)
        self.assertEqual(result, msg)
        
    def testTimeIt(self):
        msg = b''.join((b'a' for x in range(2**10)))
        response = self.client.timeit(10, msg)
        yield response
        self.assertTrue(response.locked_time >= 0)
        self.assertTrue(response.total_time >= response.locked_time)
        self.assertEqual(response.num_failures, 0)
        
    def test_multi(self):
        c = self.client
        result = yield multi_async((c.request(b'ciao'),
                                    c.request(b'pippo'),
                                    c.request(b'foo')))
        self.assertEqual(len(result), 3)
        self.assertTrue(b'ciao' in result)
        self.assertTrue(b'pippo' in result)
        self.assertTrue(b'foo' in result)
        
    def test_client(self):
        c = self.client
        yield self.test_multi()
        self.assertTrue(len(c.connection_pools), 1)
        self.assertTrue(c.available_connections)
        

@dont_run_with_thread
class TestEchoServerProcess(TestEchoServerThread):
    concurrency = 'process'


import errno
import os
import tempfile

__all__ = ['Pidfile']

class Pidfile(object):
    """\
    Manage a PID file. If a specific name is provided
    it and '"%s.oldpid" % name' will be used. Otherwise
    we create a temp file using os.mkstemp.
    """
    def __init__(self, fname=None):
        self.fname = fname
        self.pid = None
        
    def create(self, pid=None):
        pid = pid or os.getpid()
        oldpid = self.read()
        if oldpid:
            if oldpid == pid:
                return
            raise RuntimeError("Already running on PID %s " \
                               "(or pid file '%s' is stale)" %\
                                (oldpid, self.fname))
        self.pid = pid
        # Write pidfile
        if self.fname:
            fdir = os.path.dirname(self.fname)
            if fdir and not os.path.isdir(fdir):
                raise RuntimeError("%s doesn't exist. Can't create pidfile." % fdir)
        else:
            self.fname = tempfile.mktemp()
        with open(self.fname, 'w') as f:
            f.write("%s\n" % self.pid)
        # set permissions to -rw-r--r-- 
        os.chmod(self.fname, 420)
        
    def rename(self, path):
        self.unlink()
        self.fname = path
        self.create(self.pid)
        
    def unlink(self):
        """ delete pidfile"""
        try:
            with open(self.fname, "r") as f:
                pid1 =  int(f.read() or 0)
            if pid1 == self.pid:
                os.unlink(self.fname)
        except Exception:
            pass
       
    def read(self):
        """ Validate pidfile and make it stale if needed"""
        if not self.fname:
            return
        try:
            with open(self.fname, "r") as f:
                wpid = int(f.read() or 0)
                if wpid <= 0:
                    return
                return wpid
        except IOError:
            return



from base64 import b64encode, b64decode

import pulsar
from pulsar.apps.ws import WebSocketProtocol, WS
from pulsar.utils.httpurl import parse_dict_header
from pulsar.utils.pep import native_str
from pulsar.utils.websocket import FrameParser


__all__ = ['WebSocketResponse']
    
    
class WebSocketResponse(WebSocketProtocol):
    
    def __init__(self, connection, handshake, parser=None):
        # keep a reference to the websocket
        self.handshake = handshake
        handshake.finished(self)
        super(WebSocketResponse, self).__init__(connection)
        connection.set_timeout(0)
        self.parser = parser or FrameParser(kind=1)
        self.handler = handshake.current_request.websocket_handler
        if not self.handler:
            self.handler = WS()
        



import re
import logging

from pulsar.apps import pubsub
from pulsar import send, command
from pulsar.utils.pep import itervalues, iteritems


LOGGER = logging.getLogger('pulsar.pubsub')
        
        
class PubSubBackend(pubsub.PubSubBackend):
    '''Implements :class:`PubSub` in pulsar.
    
.. attribute:: monitor

    The name of :class:`pulsar.Monitor` (application) which broadcast messages
    to its workers.
'''
    def publish(self, channel, message):
        return send(self.name, 'pubsub_publish', self.id, channel, message)
        
    def subscribe(self, *channels):
        return send(self.name, 'pubsub_subscribe', self.id, *channels)
        
    def unsubscribe(self, *channels):
        return send(self.name, 'pubsub_unsubscribe', self.id, *channels)


@command()
def pubsub_publish(request, id, channel, message):
    monitor = request.actor
    if not channel or channel == '*':
        matched = ((c, reg[1]) for c, reg in iteritems(_get_pubsub_channels(monitor)))
    else:
        matched = _channel_groups(monitor, channel)
    clients = set()
    # loop over matched channels
    for channel, group in matched:
        for aid, pid in group:
            # if the id is matched we have a client
            if pid == id:
                clients.add(aid)
                monitor.send(aid, 'pubsub_broadcast', id, channel, message)
    return len(clients)
        
@command()
def pubsub_subscribe(request, id, *channels):
    client = (request.caller.aid, id)
    for channel in channels:
        _, group = _get_channel(request.actor, channel)
        group.add(client)
    return _channels_for_client(request.actor, client)
        
@command()
def pubsub_unsubscribe(request, id, *channels):
    client = (request.caller.aid, id)
    pubsub_channels = _get_pubsub_channels(request.actor)
    channels = channels or list(pubsub_channels)
    for channel in channels:
        if channel in pubsub_channels:
            _, group =  pubsub_channels[channel]
            group.discard(client)
            if not group:
                pubsub_channels.pop(channel)
    return _channels_for_client(request.actor, client)
    
@command(ack=False)
def pubsub_broadcast(request, id, channels, message):
    '''In the actor domain'''
    pubsub = PubSubBackend.get(id, actor=request.actor)
    if pubsub:
        pubsub.broadcast(channels, message)
    else:
        LOGGER.warning('Pubsub backend not available in %s', request.actor)
    
def _get_channel(monitor, channel):
    pubsub = _get_pubsub_channels(monitor)
    if channel not in pubsub:                         
        if '*' in channel:
            channel_re = re.compile(channel)
        else:
            channel_re = channel
        pubsub[channel] = (channel_re, set())
    return pubsub[channel]

def _get_pubsub_channels(actor):
    if 'pubsub_channels' not in actor.params:
        actor.params.pubsub_channels = {}
    return actor.params.pubsub_channels

def _channel_groups(actor, channel):
    for channel_re, group in itervalues(_get_pubsub_channels(actor)):
        if hasattr(channel_re, 'match'):
            g = channel_re.match(channel)
            if g:
                yield g.group(), group
        elif channel_re == channel:
            yield channel, group
            
def _channels_for_client(monitor, client):
    c = 0
    for _, group in _get_pubsub_channels(monitor).values():
        if client in group:
            c += 1
    return c